# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Version-independent api tests"""

import http.client as http_client

from oslo_serialization import jsonutils
import webob

from glance.tests import functional
from glance.tests.unit import test_versions as tv


class TestApiVersions(functional.SynchronousAPIBase):
    def setUp(self, bypass_headers=True):
        super(TestApiVersions, self).setUp(bypass_headers=bypass_headers)
        # Use version negotiation pipeline for unauthenticated endpoints
        self.start_server(enable_version_negotiation=True)

    def test_version_configurations(self):
        """Test that versioning is handled properly through all channels"""
        # Use a dummy URL for href comparison since we're testing
        # in-process
        url = 'http://localhost'
        # SynchronousAPIBase sets up multiple backends, so we need
        # both flags
        expected_versions_list = tv.get_versions_list(url,
                                                      enabled_backends=True,
                                                      enabled_cache=True)

        # Verify version choices returned.
        # Access /versions directly instead of / to avoid version
        # negotiation issues
        response = self.api_get('/versions')
        self.assertEqual(http_client.OK, response.status_code)
        content = response.json

        # Compare versions by ID and status, ignoring href URLs which
        # may differ
        actual_versions = content['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        # Create a dict for easier lookup by version ID
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id,
                          "Version %s not in expected list" % actual['id'])
            expected = expected_by_id[actual['id']]
            self.assertEqual(actual['status'], expected['status'])

    def test_v2_api_configuration(self):
        # Use a dummy URL for href comparison since we're testing
        # in-process
        url = 'http://localhost'
        # SynchronousAPIBase sets up multiple backends, so we need
        # both flags
        expected_versions_list = tv.get_versions_list(url,
                                                      enabled_backends=True,
                                                      enabled_cache=True)

        # Verify version choices returned.
        # Access /versions directly instead of / to avoid version
        # negotiation issues
        response = self.api_get('/versions')
        self.assertEqual(http_client.OK, response.status_code)
        content = response.json

        # Compare versions by ID and status, ignoring href URLs which
        # may differ
        actual_versions = content['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        # Create a dict for easier lookup by version ID
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id,
                          "Version %s not in expected list" % actual['id'])
            expected = expected_by_id[actual['id']]
            self.assertEqual(actual['status'], expected['status'])


class TestApiVersionsMultistore(functional.SynchronousAPIBase):
    def setUp(self, bypass_headers=True):
        super(TestApiVersionsMultistore, self).setUp(
            bypass_headers=bypass_headers)
        # Use version negotiation pipeline for unauthenticated endpoints
        self.start_server(enable_version_negotiation=True)

    def test_version_configurations(self):
        """Test that versioning is handled properly through all channels"""
        # Use a dummy URL for href comparison since we're testing in-process
        url = 'http://localhost'
        expected_versions_list = tv.get_versions_list(url,
                                                      enabled_backends=True,
                                                      enabled_cache=True)

        # Verify version choices returned.
        # Access /versions directly instead of / to avoid version
        # negotiation issues
        response = self.api_get('/versions')
        self.assertEqual(http_client.OK, response.status_code)
        content = response.json

        # Compare versions by ID and status, ignoring href URLs which
        # may differ
        actual_versions = content['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        for actual, expected in zip(actual_versions, expected_versions_list):
            self.assertEqual(actual['id'], expected['id'])
            self.assertEqual(actual['status'], expected['status'])

    def test_v2_api_configuration(self):
        # Use a dummy URL for href comparison since we're testing
        # in-process
        url = 'http://localhost'
        expected_versions_list = tv.get_versions_list(url,
                                                      enabled_backends=True,
                                                      enabled_cache=True)

        # Verify version choices returned.
        # Access /versions directly instead of / to avoid version
        # negotiation issues
        response = self.api_get('/versions')
        self.assertEqual(http_client.OK, response.status_code)
        content = response.json

        # Compare versions by ID and status, ignoring href URLs which
        # may differ
        actual_versions = content['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        for actual, expected in zip(actual_versions, expected_versions_list):
            self.assertEqual(actual['id'], expected['id'])
            self.assertEqual(actual['status'], expected['status'])


class TestApiPaths(functional.SynchronousAPIBase):
    def setUp(self, bypass_headers=True):
        super(TestApiPaths, self).setUp(bypass_headers=bypass_headers)
        # Use version negotiation pipeline for unauthenticated endpoints
        self.start_server(enable_version_negotiation=True)

        # Use a dummy URL for href comparison since we're testing
        # in-process
        url = 'http://localhost'
        # SynchronousAPIBase sets up multiple backends, so we need
        # both flags
        self.versions = {'versions': tv.get_versions_list(
            url, enabled_backends=True, enabled_cache=True)}
        images = {'images': []}
        self.images_json = jsonutils.dumps(images)

    def test_get_root_path(self):
        """Assert GET / with `no Accept:` header.
        Verify version choices returned.
        Bug lp:803260  no Accept header causes a 500 in glance-api
        """
        # Create a request to test with
        req = webob.Request.blank('/')
        response = self._call_api(req)
        self.assertEqual(http_client.MULTIPLE_CHOICES, response.status_code)
        content = response.json
        # Compare versions by ID and status, ignoring href URLs
        actual_versions = content['versions']
        expected_versions_list = self.versions['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id)
            self.assertEqual(
                actual['status'], expected_by_id[actual['id']]['status'])

    def test_get_root_path_with_unknown_header(self):
        """Assert GET / with Accept: unknown header
        Verify version choices returned. Verify message in API log about
        unknown accept header.
        """
        headers = {'Accept': 'unknown'}
        req = webob.Request.blank('/', headers=headers)
        response = self._call_api(req)
        self.assertEqual(http_client.MULTIPLE_CHOICES, response.status_code)
        content = response.json
        # Compare versions by ID and status, ignoring href URLs
        actual_versions = content['versions']
        expected_versions_list = self.versions['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id)
            self.assertEqual(
                actual['status'], expected_by_id[actual['id']]['status'])

    def test_get_va1_images_path(self):
        """Assert GET /va.1/images with no Accept: header
        Verify version choices returned
        """
        req = webob.Request.blank('/va.1/images')
        response = self._call_api(req)
        self.assertEqual(http_client.MULTIPLE_CHOICES, response.status_code)
        content = response.json
        # Compare versions by ID and status, ignoring href URLs
        actual_versions = content['versions']
        expected_versions_list = self.versions['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id)
            self.assertEqual(
                actual['status'], expected_by_id[actual['id']]['status'])

    def test_get_versions_path(self):
        """Assert GET /versions with no Accept: header
        Verify version choices returned
        """
        response = self.api_get('/versions')
        self.assertEqual(http_client.OK, response.status_code)
        content = response.json
        # Compare versions by ID and status, ignoring href URLs
        actual_versions = content['versions']
        expected_versions_list = self.versions['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id)
            self.assertEqual(
                actual['status'], expected_by_id[actual['id']]['status'])

    def test_get_versions_choices(self):
        """Verify version choices returned"""
        req = webob.Request.blank('/v10')
        response = self._call_api(req)
        self.assertEqual(http_client.MULTIPLE_CHOICES, response.status_code)
        content = response.json
        # Compare versions by ID and status, ignoring href URLs
        actual_versions = content['versions']
        expected_versions_list = self.versions['versions']
        self.assertEqual(len(actual_versions), len(expected_versions_list))
        expected_by_id = {v['id']: v for v in expected_versions_list}
        for actual in actual_versions:
            self.assertIn(actual['id'], expected_by_id)
            self.assertEqual(
                actual['status'], expected_by_id[actual['id']]['status'])



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Glance Image Cache Pruner

This is meant to be run as a periodic task, perhaps every half-hour.
"""

import os
import sys

from oslo_log import log as logging

# If ../glance/__init__.py exists, add ../ to Python search path, so that
# it will override what happens to be installed in /usr/(local/)lib/python...
possible_topdir = os.path.normpath(os.path.join(os.path.abspath(sys.argv[0]),
                                   os.pardir,
                                   os.pardir))
if os.path.exists(os.path.join(possible_topdir, 'glance', '__init__.py')):
    sys.path.insert(0, possible_topdir)

from glance.common import config
from glance.image_cache import pruner

CONF = config.CONF
logging.register_options(CONF)
CONF.set_default(name='use_stderr', default=True)


def main():
    print('The Glance cache pruner has been deprecated since the Epoxy '
          '(2025.1) release. It will be removed in the G (2026.1) release. '
          'The Cache API should be used instead.',
          file=sys.stderr)
    try:
        config.parse_cache_args()
        logging.setup(CONF, 'glance')

        app = pruner.Pruner()
        app.run()
    except RuntimeError as e:
        sys.exit("ERROR: %s" % e)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Functional test case that tests logging output"""

import http.client as http
import logging
import logging.handlers
import os
import stat

from glance.tests import functional


class TestLogging(functional.SynchronousAPIBase):

    """Functional tests for Glance's logging output"""

    def setUp(self):
        super(TestLogging, self).setUp()
        # Set up log file path
        self.log_file = os.path.join(self.test_dir, "api.log")
        self.file_handler = None

    def tearDown(self):
        # Clean up file handler if it was added
        if self.file_handler:
            glance_logger = logging.getLogger('glance')
            glance_logger.removeHandler(self.file_handler)
            self.file_handler.close()
        super(TestLogging, self).tearDown()

    def _setup_file_logging(self, debug=False, use_watched=False):
        """Set up file logging for the glance logger."""
        glance_logger = logging.getLogger('glance')
        # Remove existing file handler if any
        if self.file_handler:
            glance_logger.removeHandler(self.file_handler)
            self.file_handler.close()

        # Create a new file handler (WatchedFileHandler for rotation support)
        if use_watched:
            self.file_handler = logging.handlers.WatchedFileHandler(
                self.log_file)
        else:
            self.file_handler = logging.FileHandler(self.log_file)
        formatter = logging.Formatter(
            '%(asctime)s %(levelname)s [%(name)s] %(message)s')
        self.file_handler.setFormatter(formatter)
        if debug:
            self.file_handler.setLevel(logging.DEBUG)
            glance_logger.setLevel(logging.DEBUG)
        else:
            self.file_handler.setLevel(logging.INFO)
            glance_logger.setLevel(logging.INFO)
        glance_logger.addHandler(self.file_handler)

    def test_debug(self):
        """
        Test logging output proper when debug is on.
        """
        # Configure logging to write to file with debug enabled
        self.config(debug=True)
        self._setup_file_logging(debug=True)
        self.start_server()

        # Make an API call to generate log output
        response = self.api_get('/v2/images')
        self.assertEqual(http.OK, response.status_code)

        # Flush the handler to ensure all logs are written
        self.file_handler.flush()

        # Verify that debug statements appear in the API logs
        self.assertTrue(os.path.exists(self.log_file))

        with open(self.log_file, 'r') as f:
            api_log_out = f.read()

        self.assertIn('DEBUG [glance', api_log_out)

    def test_no_debug(self):
        """
        Test logging output proper when debug is off.
        """
        # Configure logging to write to file with debug disabled
        self.config(debug=False)
        self._setup_file_logging(debug=False)
        self.start_server()

        # Make an API call to generate log output
        response = self.api_get('/v2/images')
        self.assertEqual(http.OK, response.status_code)

        # Flush the handler to ensure all logs are written
        self.file_handler.flush()

        # Verify that debug statements do not appear in the API logs
        self.assertTrue(os.path.exists(self.log_file))

        with open(self.log_file, 'r') as f:
            api_log_out = f.read()

        self.assertNotIn('DEBUG [glance', api_log_out)

    def assertNotEmptyFile(self, path):
        self.assertTrue(os.path.exists(path))
        self.assertNotEqual(os.stat(path)[stat.ST_SIZE], 0)

    def test_logrotate(self):
        """
        Test that we notice when our log file has been rotated
        """
        # Configure logging to write to file using WatchedFileHandler
        # which can detect log rotation
        self.config(debug=True)
        self._setup_file_logging(debug=True, use_watched=True)
        self.start_server()

        # Make an API call to generate log output
        response = self.api_get('/v2/images')
        self.assertEqual(http.OK, response.status_code)

        # Flush the handler to ensure all logs are written
        self.file_handler.flush()

        # Verify log file exists and is not empty
        self.assertNotEmptyFile(self.log_file)

        # Rotate the log file
        os.rename(self.log_file, self.log_file + ".1")

        # Force a log message to trigger WatchedFileHandler to check
        # if the file was rotated and create a new one
        glance_logger = logging.getLogger('glance')
        glance_logger.info("Test log message after rotation")

        # Make another API call - WatchedFileHandler should detect
        # the rotation and create a new log file
        response = self.api_get('/v2/images')
        self.assertEqual(http.OK, response.status_code)

        # Flush the handler
        self.file_handler.flush()

        # Verify a new log file was created
        self.assertNotEmptyFile(self.log_file)



# Copyright 2011 OpenStack Foundation
# Copyright 2015 Mirantis, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import ipaddress
import socket
import tempfile
from unittest import mock
from urllib.parse import urlparse

import glance_store as store
from glance_store._drivers import cinder
from oslo_config import cfg
from oslo_log import log as logging
import webob

from glance.common import exception
from glance.common import store_utils
from glance.common import utils
from glance.tests.unit import base
import glance.tests.unit.utils as unit_test_utils
from glance.tests import utils as test_utils

CONF = cfg.CONF

BASE_URI = unit_test_utils.BASE_URI


class TestStoreUtils(test_utils.BaseTestCase):
    """Test glance.common.store_utils module"""

    def _test_update_store_in_location(self, metadata, store_id, expected,
                                       store_id_call_count=1,
                                       save_call_count=1):
        image = mock.Mock()
        image_repo = mock.Mock()
        image_repo.save = mock.Mock()
        context = mock.Mock()
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': metadata
        }]
        image.locations = locations
        with mock.patch.object(
                store_utils, '_get_store_id_from_uri') as mock_get_store_id:
            mock_get_store_id.return_value = store_id
            store_utils.update_store_in_locations(context, image, image_repo)
            self.assertEqual(image.locations[0]['metadata'].get(
                'store'), expected)
            self.assertEqual(store_id_call_count, mock_get_store_id.call_count)
            self.assertEqual(save_call_count, image_repo.save.call_count)

    def test_update_store_location_with_no_store(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        self._test_update_store_in_location({}, 'rbd1', 'rbd1')

    def test_update_store_location_with_different_store(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        self._test_update_store_in_location(
            {'store': 'rbd2'}, 'ceph1', 'ceph1')

    def test_update_store_location_with_same_store(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        self._test_update_store_in_location({'store': 'rbd1'}, 'rbd1', 'rbd1',
                                            store_id_call_count=0,
                                            save_call_count=0)

    def test_update_store_location_with_store_none(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        self._test_update_store_in_location({}, None, None,
                                            save_call_count=0)


class TestCinderStoreUtils(base.MultiStoreClearingUnitTest):
    """Test glance.common.store_utils module for cinder multistore"""

    @mock.patch.object(cinder.Store, 'is_image_associated_with_store')
    @mock.patch.object(cinder.Store, 'url_prefix',
                       new_callable=mock.PropertyMock)
    def _test_update_cinder_store_in_location(self, mock_url_prefix,
                                              mock_associate_store,
                                              is_valid=True,
                                              modify_source_url=False):
        volume_id = 'db457a25-8f16-4b2c-a644-eae8d17fe224'
        store_id = 'fast-cinder'
        expected = 'fast-cinder'
        image = mock.Mock()
        image_repo = mock.Mock()
        image_repo.save = mock.Mock()
        context = mock.Mock()
        mock_associate_store.return_value = is_valid
        locations = [{
            'url': 'cinder://%s' % volume_id,
            'metadata': {}
        }]
        mock_url_prefix.return_value = 'cinder://%s' % store_id
        image.locations = locations
        if modify_source_url:
            updated_location = store_utils.get_updated_store_location(
                locations, context=context)
        else:
            store_utils.update_store_in_locations(context, image, image_repo)

        if is_valid:
            # This is the case where we found an image that has an
            # old-style URL which does not include the store name,
            # but for which we know the corresponding store that
            # refers to the volume type that backs it. We expect that
            # the URL should be updated to point to the store/volume from
            # just a naked pointer to the volume, as was the old
            # format i.e. this is the case when store is valid and location
            # url, metadata are updated and image_repo.save is called
            expected_url = mock_url_prefix.return_value + '/' + volume_id
            if modify_source_url:
                # This is the case where location url is modified to be
                # compatible with multistore as below,
                # `cinder://store_id/volume_id` to avoid InvalidLocation error
                self.assertEqual(expected_url, updated_location[0].get('url'))
                self.assertEqual(expected,
                                 updated_location[0]['metadata'].get('store'))
            else:
                self.assertEqual(expected_url, image.locations[0].get('url'))
                self.assertEqual(expected, image.locations[0]['metadata'].get(
                    'store'))
                self.assertEqual(1, image_repo.save.call_count)
        else:
            # Here, we've got an image backed by a volume which does
            # not have a corresponding store specifying the volume_type.
            # Expect that we leave these alone and do not touch the
            # location URL since we cannot update it with a valid store i.e.
            # this is the case when store is invalid and location url,
            # metadata are not updated and image_repo.save is not called
            self.assertEqual(locations[0]['url'],
                             image.locations[0].get('url'))
            self.assertEqual({}, image.locations[0]['metadata'])
            self.assertEqual(0, image_repo.save.call_count)

    def test_update_cinder_store_location_valid_type(self):
        self._test_update_cinder_store_in_location()

    def test_update_cinder_store_location_invalid_type(self):
        self._test_update_cinder_store_in_location(is_valid=False)

    def test_get_updated_cinder_store_location(self):
        """
        Test if location url is modified to be compatible
        with multistore.
        """

        self._test_update_cinder_store_in_location(
            modify_source_url=True)


class TestUtils(test_utils.BaseTestCase):
    """Test routines in glance.utils"""

    def test_sort_image_locations_multistore_disabled(self):
        self.config(enabled_backends=None)
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': {'store': 'rbd1'}
        }, {
            'url': 'rbd://bbbbbbbb/images/id',
            'metadata': {'store': 'rbd2'}
        }, {
            'url': 'rbd://cccccccc/images/id',
            'metadata': {'store': 'rbd3'}
        }]
        mp = "glance.common.utils.glance_store.get_store_weight"
        with mock.patch(mp) as mock_get_store:
            utils.sort_image_locations(locations)

        # Since multistore is not enabled, it will not sort the locations
        self.assertEqual(0, mock_get_store.call_count)

    def test_sort_image_locations(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd",
            "rbd3": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="rbd1", group="glance_store")
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': {'store': 'rbd1'}
        }, {
            'url': 'rbd://bbbbbbbb/images/id',
            'metadata': {'store': 'rbd2'}
        }, {
            'url': 'rbd://cccccccc/images/id',
            'metadata': {'store': 'rbd3'}
        }]
        mp = "glance.common.utils.glance_store.get_store_weight"
        with mock.patch(mp) as mock_get_store:
            mock_get_store.return_value = 100
            utils.sort_image_locations(locations)

        # Since 3 stores are configured, internal method will be called 3 times
        self.assertEqual(3, mock_get_store.call_count)

    def test_sort_image_locations_without_metadata(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd",
            "rbd3": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="rbd1", group="glance_store")
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': {}
        }, {
            'url': 'rbd://bbbbbbbb/images/id',
            'metadata': {}
        }, {
            'url': 'rbd://cccccccc/images/id',
            'metadata': {}
        }]
        mp = "glance.common.utils.glance_store.get_store_weight"
        with mock.patch(mp) as mock_get_store:
            utils.sort_image_locations(locations)

        # Since 3 stores are configured, without store in metadata the internal
        # method will be called 0 times
        self.assertEqual(0, mock_get_store.call_count)

    def test_sort_image_locations_with_partial_metadata(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd",
            "rbd3": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="rbd1", group="glance_store")
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': {'store': 'rbd1'}
        }, {
            'url': 'rbd://bbbbbbbb/images/id',
            'metadata': {}
        }, {
            'url': 'rbd://cccccccc/images/id',
            'metadata': {}
        }]
        mp = "glance.common.utils.glance_store.get_store_weight"
        with mock.patch(mp) as mock_get_store:
            mock_get_store.return_value = 100
            utils.sort_image_locations(locations)

        # Since 3 stores are configured, but only one location has
        # store in metadata the internal
        # method will be called 1 time only
        self.assertEqual(1, mock_get_store.call_count)

    def test_sort_image_locations_unknownscheme(self):
        enabled_backends = {
            "rbd1": "rbd",
            "rbd2": "rbd",
            "rbd3": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="rbd1", group="glance_store")
        locations = [{
            'url': 'rbd://aaaaaaaa/images/id',
            'metadata': {'store': 'rbd1'}
        }, {
            'url': 'rbd://bbbbbbbb/images/id',
            'metadata': {'store': 'rbd2'}
        }, {
            'url': 'rbd://cccccccc/images/id',
            'metadata': {'store': 'rbd3'}
        }]
        mp = "glance.common.utils.glance_store.get_store_weight"
        with mock.patch(mp) as mock_get_store:
            mock_get_store.side_effect = store.UnknownScheme()
            sorted_locations = utils.sort_image_locations(locations)

        # Even though UnknownScheme exception is raised, processing continues
        self.assertEqual(3, mock_get_store.call_count)
        # Since we return 0 weight, original location order should be preserved
        self.assertEqual(locations, sorted_locations)

    def test_cooperative_reader(self):
        """Ensure cooperative reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            for chunk in utils.CooperativeReader(tmp_fd):
                bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

        bytes_read = 0
        with tempfile.TemporaryFile('w+') as tmp_fd:
            tmp_fd.write('*' * BYTES)
            tmp_fd.seek(0)
            reader = utils.CooperativeReader(tmp_fd)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_cooperative_reader_of_iterator(self):
        """Ensure cooperative reader supports iterator backends too"""
        data = b'abcdefgh'
        data_list = [data[i:i + 1] * 3 for i in range(len(data))]
        reader = utils.CooperativeReader(data_list)
        chunks = []
        while True:
            chunks.append(reader.read(3))
            if chunks[-1] == b'':
                break
        meat = b''.join(chunks)
        self.assertEqual(b'aaabbbcccdddeeefffggghhh', meat)

    def test_cooperative_reader_unbounded_read_on_iterator(self):
        """Ensure cooperative reader is happy with empty iterators"""
        data = b'abcdefgh'
        data_list = [data[i:i + 1] * 3 for i in range(len(data))]
        reader = utils.CooperativeReader(data_list)
        self.assertEqual(
            [chunk for chunk in iter(lambda: reader.read(), b'')],
            [b'aaa', b'bbb', b'ccc', b'ddd', b'eee', b'fff', b'ggg', b'hhh'])

    def test_cooperative_reader_on_iterator_with_buffer(self):
        """Ensure cooperative reader is happy with empty iterators"""
        data_list = [b'abcd', b'efgh']
        reader = utils.CooperativeReader(data_list)
        # read from part of a chunk, get the first item into the buffer
        self.assertEqual(b'ab', reader.read(2))
        # read purely from buffer
        self.assertEqual(b'c', reader.read(1))
        # unbounded read grabs the rest of the buffer
        self.assertEqual(b'd', reader.read())
        # then the whole next chunk
        self.assertEqual(b'efgh', reader.read())
        # past that, it's always empty
        self.assertEqual(b'', reader.read())

    def test_cooperative_reader_unbounded_read_on_empty_iterator(self):
        """Ensure cooperative reader is happy with empty iterators"""
        reader = utils.CooperativeReader([])
        self.assertEqual(b'', reader.read())

    def _create_generator(self, chunk_size, max_iterations):
        chars = b'abc'
        iteration = 0
        while True:
            index = iteration % len(chars)
            chunk = chars[index:index + 1] * chunk_size
            yield chunk
            iteration += 1
            if iteration >= max_iterations:
                return

    def _test_reader_chunked(self, chunk_size, read_size, max_iterations=5):
        generator = self._create_generator(chunk_size, max_iterations)
        reader = utils.CooperativeReader(generator)
        result = bytearray()
        while True:
            data = reader.read(read_size)
            if len(data) == 0:
                break
            self.assertLessEqual(len(data), read_size)
            result += data
        expected = (b'a' * chunk_size +
                    b'b' * chunk_size +
                    b'c' * chunk_size +
                    b'a' * chunk_size +
                    b'b' * chunk_size)
        self.assertEqual(expected, bytes(result))

    def test_cooperative_reader_preserves_size_chunk_less_then_read(self):
        self._test_reader_chunked(43, 101)

    def test_cooperative_reader_preserves_size_chunk_equals_read(self):
        self._test_reader_chunked(1024, 1024)

    def test_cooperative_reader_preserves_size_chunk_more_then_read(self):
        chunk_size = 16 * 1024 * 1024  # 16 Mb, as in remote http source
        read_size = 8 * 1024           # 8k, as in httplib
        self._test_reader_chunked(chunk_size, read_size)

    def test_limiting_reader(self):
        """Ensure limiting reader class accesses all bytes of file"""
        BYTES = 1024
        bytes_read = 0
        data = io.StringIO("*" * BYTES)
        for chunk in utils.LimitingReader(data, BYTES):
            bytes_read += len(chunk)

        self.assertEqual(BYTES, bytes_read)

        bytes_read = 0
        data = io.StringIO("*" * BYTES)
        reader = utils.LimitingReader(data, BYTES)
        byte = reader.read(1)
        while len(byte) != 0:
            bytes_read += 1
            byte = reader.read(1)

        self.assertEqual(BYTES, bytes_read)

    def test_limiting_reader_fails(self):
        """Ensure limiting reader class throws exceptions if limit exceeded"""
        BYTES = 1024

        def _consume_all_iter():
            bytes_read = 0
            data = io.StringIO("*" * BYTES)
            for chunk in utils.LimitingReader(data, BYTES - 1):
                bytes_read += len(chunk)

        self.assertRaises(exception.ImageSizeLimitExceeded, _consume_all_iter)

        def _consume_all_read():
            bytes_read = 0
            data = io.StringIO("*" * BYTES)
            reader = utils.LimitingReader(data, BYTES - 1)
            byte = reader.read(1)
            while len(byte) != 0:
                bytes_read += 1
                byte = reader.read(1)

        self.assertRaises(exception.ImageSizeLimitExceeded, _consume_all_read)

    def test_get_meta_from_headers(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-name": 'test',
                        'x-image-meta-virtual-size': 80}
        result = utils.get_image_meta_from_headers(resp)
        self.assertEqual({'name': 'test', 'properties': {},
                          'virtual_size': 80}, result)

    def test_get_meta_from_headers_none_virtual_size(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-name": 'test',
                        'x-image-meta-virtual-size': 'None'}
        result = utils.get_image_meta_from_headers(resp)
        self.assertEqual({'name': 'test', 'properties': {},
                          'virtual_size': None}, result)

    def test_get_meta_from_headers_bad_headers(self):
        resp = webob.Response()
        resp.headers = {"x-image-meta-bad": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)
        resp.headers = {"x-image-meta-": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)
        resp.headers = {"x-image-meta-*": 'test'}
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_image_meta_from_headers, resp)

    def test_image_meta(self):
        image_meta = {'x-image-meta-size': 'test'}
        image_meta_properties = {'properties': {'test': "test"}}
        actual = utils.image_meta_to_http_headers(image_meta)
        actual_test2 = utils.image_meta_to_http_headers(
            image_meta_properties)
        self.assertEqual({'x-image-meta-x-image-meta-size': 'test'}, actual)
        self.assertEqual({'x-image-meta-property-test': 'test'},
                         actual_test2)

    def test_create_mashup_dict_with_different_core_custom_properties(self):
        image_meta = {
            'id': 'test-123',
            'name': 'fake_image',
            'status': 'active',
            'created_at': '',
            'min_disk': '10G',
            'min_ram': '1024M',
            'protected': False,
            'locations': '',
            'checksum': 'c1234',
            'owner': '',
            'disk_format': 'raw',
            'container_format': 'bare',
            'size': '123456789',
            'virtual_size': '123456789',
            'is_public': 'public',
            'deleted': True,
            'updated_at': '',
            'properties': {'test_key': 'test_1234'},
        }

        mashup_dict = utils.create_mashup_dict(image_meta)
        self.assertNotIn('properties', mashup_dict)
        self.assertEqual(image_meta['properties']['test_key'],
                         mashup_dict['test_key'])

    def test_create_mashup_dict_with_same_core_custom_properties(self):
        image_meta = {
            'id': 'test-123',
            'name': 'fake_image',
            'status': 'active',
            'created_at': '',
            'min_disk': '10G',
            'min_ram': '1024M',
            'protected': False,
            'locations': '',
            'checksum': 'c1234',
            'owner': '',
            'disk_format': 'raw',
            'container_format': 'bare',
            'size': '123456789',
            'virtual_size': '123456789',
            'is_public': 'public',
            'deleted': True,
            'updated_at': '',
            'properties': {'min_ram': '2048M'},
        }

        mashup_dict = utils.create_mashup_dict(image_meta)
        self.assertNotIn('properties', mashup_dict)
        self.assertNotEqual(image_meta['properties']['min_ram'],
                            mashup_dict['min_ram'])
        self.assertEqual(image_meta['min_ram'], mashup_dict['min_ram'])

    def test_mutating(self):
        class FakeContext(object):
            def __init__(self):
                self.read_only = False

        class Fake(object):
            def __init__(self):
                self.context = FakeContext()

        def fake_function(req, context):
            return 'test passed'

        req = webob.Request.blank('/some_request')
        result = utils.mutating(fake_function)
        self.assertEqual("test passed", result(req, Fake()))

    def test_valid_hostname(self):
        valid_inputs = ['localhost',
                        'glance04-a'
                        'G',
                        '528491']

        for input_str in valid_inputs:
            self.assertTrue(utils.is_valid_hostname(input_str))

    def test_valid_hostname_fail(self):
        invalid_inputs = ['localhost.localdomain',
                          '192.168.0.1',
                          '\u2603',
                          'glance02.stack42.local']

        for input_str in invalid_inputs:
            self.assertFalse(utils.is_valid_hostname(input_str))

    def test_valid_fqdn(self):
        valid_inputs = ['localhost.localdomain',
                        'glance02.stack42.local'
                        'glance04-a.stack47.local',
                        'img83.glance.xn--penstack-r74e.org']

        for input_str in valid_inputs:
            self.assertTrue(utils.is_valid_fqdn(input_str))

    def test_valid_fqdn_fail(self):
        invalid_inputs = ['localhost',
                          '192.168.0.1',
                          '999.88.77.6',
                          '\u2603.local',
                          'glance02.stack42']

        for input_str in invalid_inputs:
            self.assertFalse(utils.is_valid_fqdn(input_str))

    def test_valid_host_port_string(self):
        valid_pairs = ['10.11.12.13:80',
                       '172.17.17.1:65535',
                       '[fe80::a:b:c:d]:9990',
                       'localhost:9990',
                       'localhost.localdomain:9990',
                       'glance02.stack42.local:1234',
                       'glance04-a.stack47.local:1234',
                       'img83.glance.xn--penstack-r74e.org:13080']

        for pair_str in valid_pairs:
            host, port = utils.parse_valid_host_port(pair_str)

            escaped = pair_str.startswith('[')
            expected_host = '%s%s%s' % ('[' if escaped else '', host,
                                        ']' if escaped else '')

            self.assertTrue(pair_str.startswith(expected_host))
            self.assertGreater(port, 0)

            expected_pair = '%s:%d' % (expected_host, port)
            self.assertEqual(expected_pair, pair_str)

    def test_valid_host_port_string_fail(self):
        invalid_pairs = ['',
                         '10.11.12.13',
                         '172.17.17.1:99999',
                         '290.12.52.80:5673',
                         'absurd inputs happen',
                         '\u2601',
                         '\u2603:8080',
                         'fe80::1',
                         '[fe80::2]',
                         '<fe80::3>:5673',
                         '[fe80::a:b:c:d]9990',
                         'fe80:a:b:c:d:e:f:1:2:3:4',
                         'fe80:a:b:c:d:e:f:g',
                         'fe80::1:8080',
                         '[fe80:a:b:c:d:e:f:g]:9090',
                         '[a:b:s:u:r:d]:fe80']

        for pair in invalid_pairs:
            self.assertRaises(ValueError,
                              utils.parse_valid_host_port,
                              pair)

    def test_get_stores_from_request_returns_default(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")

        req = webob.Request.blank('/some_request')
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = utils.get_stores_from_request(req, {})
            self.assertEqual(["ceph1"], result)
            mock_get_store.assert_called_once_with("ceph1")

    def test_get_stores_from_request_returns_stores_from_body(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")

        body = {"stores": ["ceph1", "ceph2"]}
        req = webob.Request.blank("/some_request")
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = utils.get_stores_from_request(req, body)
            self.assertEqual(["ceph1", "ceph2"], result)
            mock_get_store.assert_any_call("ceph1")
            mock_get_store.assert_any_call("ceph2")
            self.assertEqual(mock_get_store.call_count, 2)

    def test_get_stores_from_request_returns_store_from_headers(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")

        headers = {"x-image-meta-store": "ceph2"}
        req = webob.Request.blank("/some_request", headers=headers)
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = utils.get_stores_from_request(req, {})
            self.assertEqual(["ceph2"], result)
            mock_get_store.assert_called_once_with("ceph2")

    def test_get_stores_from_request_raises_bad_request(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")
        headers = {"x-image-meta-store": "ceph2"}
        body = {"stores": ["ceph1"]}
        req = webob.Request.blank("/some_request", headers=headers)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_stores_from_request, req, body)

    def test_get_stores_from_request_returns_all_stores(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        reserved_stores = {
            'os_glance_staging_store': 'file',
            'os_glance_tasks_store': 'file'
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF, reserved_stores=reserved_stores)
        self.config(default_backend="ceph1", group="glance_store")
        body = {"all_stores": True}
        req = webob.Request.blank("/some_request")
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = sorted(utils.get_stores_from_request(req, body))
            self.assertEqual(["ceph1", "ceph2"], result)
            mock_get_store.assert_any_call("ceph1")
            mock_get_store.assert_any_call("ceph2")
            self.assertEqual(mock_get_store.call_count, 2)
            self.assertNotIn('os_glance_staging_store', result)
            self.assertNotIn('os_glance_tasks_store', result)

    def test_get_stores_from_request_excludes_reserved_stores(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")
        body = {"all_stores": True}
        req = webob.Request.blank("/some_request")
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = sorted(utils.get_stores_from_request(req, body))
            self.assertEqual(["ceph1", "ceph2"], result)
            mock_get_store.assert_any_call("ceph1")
            mock_get_store.assert_any_call("ceph2")
            self.assertEqual(mock_get_store.call_count, 2)

    def test_get_stores_from_request_excludes_readonly_store(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd",
            "http": "http"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")
        body = {"all_stores": True}
        req = webob.Request.blank("/some_request")
        mp = "glance.common.utils.glance_store.get_store_from_store_identifier"
        with mock.patch(mp) as mock_get_store:
            result = sorted(utils.get_stores_from_request(req, body))
            self.assertNotIn("http", result)
            self.assertEqual(["ceph1", "ceph2"], result)
            mock_get_store.assert_any_call("ceph1")
            mock_get_store.assert_any_call("ceph2")
            self.assertEqual(mock_get_store.call_count, 2)

    def test_get_stores_from_request_raises_bad_request_with_all_stores(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")
        headers = {"x-image-meta-store": "ceph2"}
        body = {"stores": ["ceph1"], "all_stores": True}
        req = webob.Request.blank("/some_request", headers=headers)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          utils.get_stores_from_request, req, body)

    def test_single_store_http_enabled_and_http_not_in_url(self):
        self.config(stores="http,file", group="glance_store")
        loc_url = "rbd://aaaaaaaa/images/id"
        self.assertFalse(utils.is_http_store_configured(loc_url))

    def test_single_store_http_disabled_and_http_in_url(self):
        self.config(stores="rbd,file", group="glance_store")
        loc_url = BASE_URI
        self.assertFalse(utils.is_http_store_configured(loc_url))

    def test_single_store_http_enabled_and_http_in_url(self):
        self.config(stores="http,file", group="glance_store")
        loc_url = BASE_URI
        self.assertTrue(utils.is_http_store_configured(loc_url))

    def test_multiple_store_http_enabled_and_http_not_in_url(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd",
            "http": "http"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="http", group="glance_store")
        loc_url = "rbd://aaaaaaaa/images/id"
        self.assertFalse(utils.is_http_store_configured(loc_url))

    def test_multiple_store_http_disabled_and_http_in_url(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd",
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="ceph1", group="glance_store")
        loc_url = BASE_URI
        self.assertFalse(utils.is_http_store_configured(loc_url))

    def test_multiple_store_http_enabled_and_http_in_url(self):
        enabled_backends = {
            "ceph1": "rbd",
            "ceph2": "rbd",
            "http": "http"
        }
        self.config(enabled_backends=enabled_backends)
        store.register_store_opts(CONF)
        self.config(default_backend="http", group="glance_store")
        loc_url = BASE_URI
        self.assertTrue(utils.is_http_store_configured(loc_url))


class SplitFilterOpTestCase(test_utils.BaseTestCase):

    def test_less_than_operator(self):
        expr = 'lt:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lt', 'bar'), returned)

    def test_less_than_equal_operator(self):
        expr = 'lte:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lte', 'bar'), returned)

    def test_greater_than_operator(self):
        expr = 'gt:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('gt', 'bar'), returned)

    def test_greater_than_equal_operator(self):
        expr = 'gte:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('gte', 'bar'), returned)

    def test_not_equal_operator(self):
        expr = 'neq:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('neq', 'bar'), returned)

    def test_equal_operator(self):
        expr = 'eq:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', 'bar'), returned)

    def test_in_operator(self):
        expr = 'in:bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('in', 'bar'), returned)

    def test_split_filter_value_for_quotes(self):
        expr = '\"fake\\\"name\",fakename,\"fake,name\"'
        returned = utils.split_filter_value_for_quotes(expr)
        list_values = ['fake\\"name', 'fakename', 'fake,name']
        self.assertEqual(list_values, returned)

    def test_validate_quotes(self):
        expr = '\"aaa\\\"aa\",bb,\"cc\"'
        returned = utils.validate_quotes(expr)
        self.assertIsNone(returned)

        invalid_expr = ['\"aa', 'ss\"', 'aa\"bb\"cc', '\"aa\"\"bb\"']
        for expr in invalid_expr:
            self.assertRaises(exception.InvalidParameterValue,
                              utils.validate_quotes,
                              expr)

    def test_default_operator(self):
        expr = 'bar'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', expr), returned)

    def test_default_operator_with_datetime(self):
        expr = '2015-08-27T09:49:58Z'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('eq', expr), returned)

    def test_operator_with_datetime(self):
        expr = 'lt:2015-08-27T09:49:58Z'
        returned = utils.split_filter_op(expr)
        self.assertEqual(('lt', '2015-08-27T09:49:58Z'), returned)


class EvaluateFilterOpTestCase(test_utils.BaseTestCase):

    def test_less_than_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'lt', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'lt', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'lt', 10))

    def test_less_than_equal_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'lte', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'lte', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'lte', 10))

    def test_greater_than_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'gt', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'gt', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'gt', 10))

    def test_greater_than_equal_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'gte', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'gte', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'gte', 10))

    def test_not_equal_operator(self):
        self.assertTrue(utils.evaluate_filter_op(9, 'neq', 10))
        self.assertFalse(utils.evaluate_filter_op(10, 'neq', 10))
        self.assertTrue(utils.evaluate_filter_op(11, 'neq', 10))

    def test_equal_operator(self):
        self.assertFalse(utils.evaluate_filter_op(9, 'eq', 10))
        self.assertTrue(utils.evaluate_filter_op(10, 'eq', 10))
        self.assertFalse(utils.evaluate_filter_op(11, 'eq', 10))

    def test_invalid_operator(self):
        self.assertRaises(exception.InvalidFilterOperatorValue,
                          utils.evaluate_filter_op, '10', 'bar', '8')


class ImportURITestCase(test_utils.BaseTestCase):

    def test_validate_import_uri(self):
        self.assertTrue(utils.validate_import_uri("http://foo.com"))

        self.config(allowed_schemes=['http'],
                    group='import_filtering_opts')
        self.config(allowed_hosts=['example.com'],
                    group='import_filtering_opts')
        self.assertTrue(utils.validate_import_uri("http://example.com"))

        self.config(allowed_ports=['8080'],
                    group='import_filtering_opts')
        self.assertTrue(utils.validate_import_uri("http://example.com:8080"))

    def test_invalid_import_uri(self):
        self.assertFalse(utils.validate_import_uri(""))

        self.assertFalse(utils.validate_import_uri("fake_uri"))
        self.config(disallowed_schemes=['ftp'],
                    group='import_filtering_opts')
        self.assertFalse(utils.validate_import_uri("ftp://example.com"))

        self.config(disallowed_hosts=['foo.com'],
                    group='import_filtering_opts')
        self.assertFalse(utils.validate_import_uri("ftp://foo.com"))

        self.config(disallowed_ports=['8484'],
                    group='import_filtering_opts')
        self.assertFalse(utils.validate_import_uri("http://localhost:8484"))

    def test_ignored_filtering_options(self):
        LOG = logging.getLogger('glance.common.utils')
        with mock.patch.object(LOG, 'debug') as mock_run:
            self.config(allowed_schemes=['https', 'ftp'],
                        group='import_filtering_opts')
            self.config(disallowed_schemes=['ftp'],
                        group='import_filtering_opts')
            self.assertTrue(utils.validate_import_uri("ftp://foo.com"))
            mock_run.assert_called_once()
        with mock.patch.object(LOG, 'debug') as mock_run:
            self.config(allowed_schemes=[],
                        group='import_filtering_opts')
            self.config(disallowed_schemes=[],
                        group='import_filtering_opts')
            self.config(allowed_hosts=['example.com', 'foo.com'],
                        group='import_filtering_opts')
            self.config(disallowed_hosts=['foo.com'],
                        group='import_filtering_opts')
            self.assertTrue(utils.validate_import_uri("ftp://foo.com"))
            mock_run.assert_called_once()
        with mock.patch.object(LOG, 'debug') as mock_run:
            self.config(allowed_hosts=[],
                        group='import_filtering_opts')
            self.config(disallowed_hosts=[],
                        group='import_filtering_opts')
            self.config(allowed_ports=[8080, 8484],
                        group='import_filtering_opts')
            self.config(disallowed_ports=[8484],
                        group='import_filtering_opts')
            self.assertTrue(utils.validate_import_uri("ftp://foo.com:8484"))
            mock_run.assert_called_once()

    def test_validate_import_uri_ip_rejection(self):
        """Test that encoded IP addresses are rejected (not normalized)."""
        # Test that standard IP is blocked when in blacklist
        self.config(disallowed_hosts=['127.0.0.1'],
                    group='import_filtering_opts')
        self.config(allowed_ports=[80],
                    group='import_filtering_opts')
        self.assertFalse(utils.validate_import_uri("http://127.0.0.1:80/"))

        # Test that encoded IP (decimal) is rejected
        result = utils.validate_import_uri("http://2130706433:80/")
        self.assertFalse(result)

        # Test that shorthand IP addresses are rejected
        self.assertFalse(utils.validate_import_uri("http://127.1:80/"))
        self.assertFalse(utils.validate_import_uri("http://10.1:80/"))
        self.assertFalse(utils.validate_import_uri("http://192.168.1:80/"))

        # Test with allowed host - encoded IP should still be rejected
        self.config(disallowed_hosts=[],
                    group='import_filtering_opts')
        self.config(allowed_hosts=['127.0.0.1'],
                    group='import_filtering_opts')
        self.assertTrue(utils.validate_import_uri("http://127.0.0.1:80/"))
        self.assertFalse(utils.validate_import_uri("http://2130706433:80/"))

    @mock.patch('glance.common.utils.socket.getaddrinfo')
    def test_normalize_hostname(self, mock_getaddrinfo):
        """Test the normalize_hostname function."""
        # Test standard IPv4 - should return normalized
        result = utils.normalize_hostname("127.0.0.1")
        self.assertEqual(result, "127.0.0.1")
        mock_getaddrinfo.assert_not_called()

        # Test standard IPv4 with different format - should normalize
        result = utils.normalize_hostname("192.168.1.1")
        self.assertEqual(result, "192.168.1.1")
        mock_getaddrinfo.assert_not_called()

        # Test encoded IP (decimal) - should be rejected (fails isdigit check)
        result = utils.normalize_hostname("2130706433")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_not_called()

        # Test hex encoded IP - should be rejected (DNS resolution fails)
        mock_getaddrinfo.reset_mock()
        mock_getaddrinfo.side_effect = socket.gaierror(
            "Name or service not known")
        result = utils.normalize_hostname("0x7f000001")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_called_once_with("0x7f000001.", 80)

        # Test octal integer encoded IP - should be rejected
        # (fails isdigit check)
        mock_getaddrinfo.reset_mock()
        result = utils.normalize_hostname("017700000001")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_not_called()

        # Test octal dotted-decimal encoded IP - should be rejected
        # (all digits/dots)
        mock_getaddrinfo.reset_mock()
        result = utils.normalize_hostname("0177.0.0.01")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_not_called()

        # Test mixed octal/decimal dotted-decimal - should be rejected
        # (all digits/dots)
        mock_getaddrinfo.reset_mock()
        result = utils.normalize_hostname("0177.0.0.1")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_not_called()

        # Test IPv6 address - should normalize to standard format
        mock_getaddrinfo.reset_mock()
        result = utils.normalize_hostname("::1")
        self.assertEqual(result, "::1")
        mock_getaddrinfo.assert_not_called()

        result = utils.normalize_hostname("2001:db8::1")
        self.assertEqual(result, "2001:db8::1")
        mock_getaddrinfo.assert_not_called()

        # Test IPv6-mapped IPv4 - should normalize
        result = utils.normalize_hostname("::ffff:127.0.0.1")
        ipv6 = ipaddress.IPv6Address(result)
        expected = ipaddress.IPv6Address("::ffff:127.0.0.1")
        self.assertEqual(ipv6, expected)
        mock_getaddrinfo.assert_not_called()

        # Test shorthand IP addresses - should be rejected
        mock_getaddrinfo.reset_mock()
        result = utils.normalize_hostname("127.1")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_not_called()

        result = utils.normalize_hostname("10.1")
        self.assertIsNone(result)

        result = utils.normalize_hostname("192.168.1")
        self.assertIsNone(result)

        # Test valid hostname - should return unchanged
        mock_getaddrinfo.reset_mock()
        mock_getaddrinfo.side_effect = None
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        result = utils.normalize_hostname("example.com")
        self.assertEqual(result, "example.com")
        mock_getaddrinfo.assert_called_once_with("example.com.", 80)

        # Test valid domain starting with digit (3m.com)
        mock_getaddrinfo.reset_mock()
        mock_getaddrinfo.side_effect = None
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        result = utils.normalize_hostname("3m.com")
        self.assertEqual(result, "3m.com")
        mock_getaddrinfo.assert_called_once_with("3m.com.", 80)

        # Test valid domain starting with 0x (0xdeadbeef.com)
        mock_getaddrinfo.reset_mock()
        mock_getaddrinfo.side_effect = None
        mock_getaddrinfo.return_value = [('', '', '', '', ('', 80))]
        result = utils.normalize_hostname("0xdeadbeef.com")
        self.assertEqual(result, "0xdeadbeef.com")
        mock_getaddrinfo.assert_called_once_with(
            "0xdeadbeef.com.", 80)

        # Test invalid/unresolvable hostname - should be rejected
        mock_getaddrinfo.reset_mock()
        mock_getaddrinfo.side_effect = socket.gaierror(
            "Name or service not known")
        result = utils.normalize_hostname("invalid-hostname-12345")
        self.assertIsNone(result)
        mock_getaddrinfo.assert_called_once_with(
            "invalid-hostname-12345.", 80)

    def test_validate_import_uri_ipv6_validation(self):
        """Test IPv6 addresses are properly validated against blacklist."""
        # Test that IPv6 localhost is blocked when in blacklist
        self.config(disallowed_hosts=['::1'],
                    group='import_filtering_opts')
        self.config(allowed_ports=[80],
                    group='import_filtering_opts')
        # IPv6 addresses in URLs are in brackets, but urlparse removes them
        # So we test with the hostname directly
        self.assertFalse(
            utils.validate_import_uri("http://[::1]:80/"))

        # Test that IPv6 address not in blacklist is allowed
        self.config(disallowed_hosts=[],
                    group='import_filtering_opts')
        self.config(allowed_hosts=['2001:db8::1'],
                    group='import_filtering_opts')
        self.assertTrue(utils.validate_import_uri("http://[2001:db8::1]:80/"))

        # Test that IPv6 localhost can be blocked separately from IPv4
        # This ensures IPv6 addresses are properly normalized and can be
        # blacklisted
        self.config(disallowed_hosts=['127.0.0.1'],
                    group='import_filtering_opts')
        self.config(allowed_hosts=[],  # Explicitly clear whitelist
                    group='import_filtering_opts')
        self.config(allowed_ports=[80],
                    group='import_filtering_opts')
        # IPv6 localhost should pass if not in blacklist (no whitelist)
        # The fix ensures IPv6 is normalized and can be blacklisted separately
        result = utils.validate_import_uri("http://[::1]:80/")
        # If ::1 is not in blacklist and no whitelist, it will pass
        # Administrators should add both IPv4 and IPv6 to blacklist if needed
        self.assertTrue(result)

        # Test that IPv6 can be blacklisted separately
        self.config(disallowed_hosts=['127.0.0.1', '::1'],
                    group='import_filtering_opts')
        self.assertFalse(utils.validate_import_uri("http://[::1]:80/"))


class S3CredentialUpdateTestCase(test_utils.BaseTestCase):

    def setUp(self):
        super(S3CredentialUpdateTestCase, self).setUp()
        self.context = mock.Mock()
        self.image_repo = mock.Mock()
        self.image = mock.Mock()
        self.image.locations = []
        self.image.image_id = 'test-image-id'

    def _create_s3_location(self, url, store_name=None):
        location = {
            'url': url,
            'metadata': {}
        }
        if store_name:
            location['metadata']['store'] = store_name
        return location

    def _mock_store_config(self, stores):
        scheme_map = {}
        for store_name, credentials in stores.items():
            store_instance = mock.Mock()
            store_instance.access_key = credentials['access_key']
            store_instance.secret_key = credentials['secret_key']
            # Add bucket and s3_host for URL matching
            store_instance.bucket = credentials.get('bucket', 'test-bucket')
            store_instance.s3_host = credentials.get(
                's3_host', 's3.amazonaws.com')
            # Add url_prefix for _get_store_id_from_uri
            store_instance.url_prefix = credentials.get('url_prefix', '')
            scheme_map[store_name] = {'store': store_instance}
        return scheme_map

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_with_store_metadata_match(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test when store exists and credentials already match."""
        mock_conf.enabled_backends = ['s3_store']
        stores = {
            's3_store': {
                'access_key': 'old_key',
                'secret_key': 'old_secret',
                'bucket': 'bucket',
                's3_host': 's3.amazonaws.com'
            }
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket/object',
            's3_store')
        self.image.locations = [location]
        original_url = location['url']
        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        # Should not save since no update was needed
        self.image_repo.save.assert_not_called()
        # URL should remain unchanged since credentials already match
        self.assertEqual(location['url'], original_url)

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_with_store_metadata_no_match(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test when store exists but credentials don't match."""
        mock_conf.enabled_backends = ['s3_store']
        stores = {
            's3_store': {
                'access_key': 'new_key',
                'secret_key': 'new_secret',
            }
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        location = self._create_s3_location(
            's3://old_key:old_secret@bucket/object', 's3_store')
        self.image.locations = [location]
        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        # Verify - credentials should be updated
        expected_url = 's3://new_key:new_secret@bucket/object'
        self.assertEqual(location['url'], expected_url)

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_multi_store_no_match(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test multi-store scenario when no URL matches."""
        mock_get_store_id.return_value = None
        mock_conf.enabled_backends = ['store1', 'store2']
        stores = {
            'store1': {
                'access_key': 'key1',
                'secret_key': 'secret1',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'key2',
                'secret_key': 'secret2',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }
        # Create URL that doesn't match any store's bucket
        location = self._create_s3_location(
            's3://unknown_key:unknown_secret@s3.amazonaws.com/'
            'unknown-bucket/object')
        self.image.locations = [location]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        # Verify - no store should be added since no match found
        self.assertNotIn('store', location['metadata'])

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_same_credentials_different_buckets(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test multi-store with same credentials but different buckets."""
        mock_get_store_id.return_value = None
        mock_conf.enabled_backends = ['store1', 'store2']
        stores = {
            'store1': {
                'access_key': 'same_key',
                'secret_key': 'same_secret',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'same_key',
                'secret_key': 'same_secret',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }
        # Create URLs for both stores with old credentials
        location1 = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket1/object1',
            'store1')
        location2 = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket2/object2',
            'store2')
        self.image.locations = [location1, location2]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        # Both URLs should be updated with new credentials
        expected_url1 = (
            's3://same_key:same_secret@s3.amazonaws.com/bucket1/object1')
        expected_url2 = (
            's3://same_key:same_secret@s3.amazonaws.com/bucket2/object2')
        self.assertEqual(location1['url'], expected_url1)
        self.assertEqual(location2['url'], expected_url2)

    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    @mock.patch('glance.common.store_utils.LOG')
    def test_update_s3_credentials_unknown_scheme(
            self, mock_log, mock_conf, mock_store_api):
        """Test with unknown scheme - S3 URL but no S3 stores configured."""
        # Empty location map means no stores are configured
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {}

        location = self._create_s3_location(
            's3://key:secret@bucket/object', 's3_store')
        self.image.locations = [location]
        original_url = location['url']

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        # URL should remain unchanged
        self.assertEqual(location['url'], original_url)
        # Verify that the debug log was called for unknown scheme
        # Note: URI is not logged for S3 schemes to avoid credential exposure
        mock_log.debug.assert_called_once_with(
            "Unknown scheme '%(scheme)s' found in uri",
            {'scheme': 's3'})

    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    @mock.patch('glance.common.store_utils.LOG')
    def test_update_s3_credentials_no_matching_store(
            self, mock_log, mock_conf, mock_store_api):
        """Test when no S3 store matches the bucket."""
        # Configure stores but with different buckets
        stores = {
            'store1': {
                'access_key': 'key1',
                'secret_key': 'secret1',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'key2',
                'secret_key': 'secret2',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        # Test with bucket that doesn't match any store
        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/unknown-bucket/object')
        self.image.locations = [location]
        original_url = location['url']

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        # URL should remain unchanged
        self.assertEqual(location['url'], original_url)
        expected_calls = [
            mock.call("Invalid location uri %s",
                      's3://old_key:old_secret@s3.amazonaws.com/'
                      'unknown-bucket/object'),
            mock.call("No S3 store found for image %(image_id)s",
                      {'image_id': 'object'})
        ]
        mock_log.warning.assert_has_calls(expected_calls)

    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_multiple_locations(
            self, mock_conf, mock_store_api):
        """Test credential update with multiple S3 locations."""
        mock_conf.enabled_backends = ['s3_store']
        stores = {
            's3_store': {
                'access_key': 'new_key',
                'secret_key': 'new_secret',
            }
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }
        location1 = self._create_s3_location(
            's3://old_key1:old_secret1@bucket1/object1', 's3_store')
        location2 = self._create_s3_location(
            's3://old_key2:old_secret2@bucket2/object2', 's3_store')
        self.image.locations = [location1, location2]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        # both locations should be updated
        expected_url1 = 's3://new_key:new_secret@bucket1/object1'
        expected_url2 = 's3://new_key:new_secret@bucket2/object2'
        self.assertEqual(location1['url'], expected_url1)
        self.assertEqual(location2['url'], expected_url2)

    def test_update_s3_url_helper_function(self):
        """Test the _update_s3_url helper function directly."""
        # Test basic URL update
        url = 's3://old_key:old_secret@bucket/object'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@bucket/object'
        self.assertEqual(result, expected)

        # Test URL without existing credentials
        url = 's3://bucket/object'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@bucket/object'
        self.assertEqual(result, expected)

        # Test with s3+https scheme
        url = 's3+https://old_key:old_secret@bucket/object'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3+https://new_key:new_secret@bucket/object'
        self.assertEqual(result, expected)

    def test_update_s3_url_with_slash_in_secret(self):
        """Test _update_s3_url handles forward slash in secret key.

        AWS S3 secret access keys can contain '/'.
        urlparse incorrectly interprets '/' as path separator, so we
        use rfind('@') to correctly locate the credentials separator.
        """
        # Secret key contains forward slash
        url = 's3+https://old_key:old/secret@bucket/object'
        result = store_utils._update_s3_url(url, 'new_key', 'new/secret')
        expected = 's3+https://new_key:new/secret@bucket/object'
        self.assertEqual(result, expected)

        # Secret key contains multiple forward slashes
        url = 's3+https://KEY:a/b/c@bucket/object'
        result = store_utils._update_s3_url(url, 'NEW', 'x/y/z')
        expected = 's3+https://NEW:x/y/z@bucket/object'
        self.assertEqual(result, expected)

    def test_update_s3_url_preserves_query_and_fragment(self):
        """Test _update_s3_url preserves query parameters and fragments."""
        # URL with query parameters
        url = 's3://old_key:old_secret@host.com/bucket/obj?param=value'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@host.com/bucket/obj?param=value'
        self.assertEqual(result, expected)

        # URL with fragment
        url = 's3://old_key:old_secret@host.com/bucket/obj#fragment'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@host.com/bucket/obj#fragment'
        self.assertEqual(result, expected)

        # URL with both query and fragment
        url = 's3://old_key:old_secret@host.com/bucket/obj?p=value#frag'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@host.com/bucket/obj?p=value#frag'
        self.assertEqual(result, expected)

    def test_update_s3_url_with_port(self):
        """Test _update_s3_url handles hosts with port numbers."""
        url = 's3://old_key:old_secret@host.com:8080/bucket/obj'
        result = store_utils._update_s3_url(url, 'new_key', 'new_secret')
        expected = 's3://new_key:new_secret@host.com:8080/bucket/obj'
        self.assertEqual(result, expected)

        # Port with slash in secret
        url = 's3://old_key:old/secret@host.com:9000/bucket/obj'
        result = store_utils._update_s3_url(url, 'new_key', 'new/secret')
        expected = 's3://new_key:new/secret@host.com:9000/bucket/obj'
        self.assertEqual(result, expected)

    def test_construct_s3_url(self):
        """Test _construct_s3_url with all attributes present."""
        store_instance = mock.Mock()
        store_instance.access_key = 'key'
        store_instance.secret_key = 'secret'
        store_instance.s3_host = 's3.amazonaws.com'
        store_instance.bucket = 'bucket'

        result = store_utils._construct_s3_url(
            store_instance, 's3', '/object/path')
        expected = 's3://key:secret@s3.amazonaws.com/bucket/object/path'
        self.assertEqual(result, expected)

    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_different_object_path(
            self, mock_conf, mock_store_api):
        """Test that URLs with different object paths are updated."""
        mock_conf.enabled_backends = ['s3_store']
        stores = {
            's3_store': {
                'access_key': 'new_key',
                'secret_key': 'new_secret',
                's3_host': 's3.amazonaws.com',
                'bucket': 'bucket'
            }
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }
        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket/old-object',
            's3_store')
        self.image.locations = [location]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        expected_url = (
            's3://new_key:new_secret@s3.amazonaws.com/bucket/old-object')
        self.assertEqual(location['url'], expected_url)

    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_update_s3_credentials_mixed_schemes(
            self, mock_conf, mock_store_api):
        """Test with mixed URL schemes - only S3 should be processed."""
        mock_conf.enabled_backends = ['s3_store']
        stores = {
            's3_store': {
                'access_key': 'new_key',
                'secret_key': 'new_secret',
                's3_host': 's3.amazonaws.com',
                'bucket': 'bucket'
            }
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        s3_location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket/object',
            's3_store')
        http_location = self._create_s3_location('http://example.com/image')
        self.image.locations = [s3_location, http_location]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)
        # only S3 location should be updated
        expected_s3_url = (
            's3://new_key:new_secret@s3.amazonaws.com/bucket/object')
        self.assertEqual(s3_location['url'], expected_s3_url)
        self.assertEqual(http_location['url'], 'http://example.com/image')

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_find_store_by_bucket_legacy_single_store(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test finding store by bucket for legacy single-store images."""
        mock_conf.enabled_backends = ['store1', 'store2']
        mock_get_store_id.return_value = None  # No store found by URI
        stores = {
            'store1': {
                'access_key': 'key1',
                'secret_key': 'secret1',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'key2',
                'secret_key': 'secret2',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        # Test legacy image without store metadata
        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket1/object')
        self.image.locations = [location]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        self.assertEqual(location['metadata']['store'], 'store1')
        expected_url = 's3://key1:secret1@s3.amazonaws.com/bucket1/object'
        self.assertEqual(location['url'], expected_url)

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_find_store_by_bucket_no_match(
            self, mock_conf, mock_store_api, mock_get_store_id):
        """Test when no store matches the bucket."""
        mock_conf.enabled_backends = ['store1', 'store2']
        mock_get_store_id.return_value = None  # No store found by URI
        stores = {
            'store1': {
                'access_key': 'key1',
                'secret_key': 'secret1',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'key2',
                'secret_key': 'secret2',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        # Test with bucket that doesn't match any store
        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/unknown-bucket/object')
        self.image.locations = [location]
        original_url = location['url']

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        # Should not update anything since no store matches
        self.assertNotIn('store', location['metadata'])
        self.assertEqual(location['url'], original_url)

    def test_find_store_by_bucket_helper_function(self):
        """Test _find_store_by_bucket helper function directly."""
        store1 = mock.Mock()
        store1.bucket = 'bucket1'
        store2 = mock.Mock()
        store2.bucket = 'bucket2'
        location_map = {
            's3': {
                'store1': {'store': store1},
                'store2': {'store': store2}
            }
        }

        # Test finding store by bucket
        parsed_uri = urlparse(
            's3://key:secret@s3.amazonaws.com/bucket1/object')
        store_name, store_instance = store_utils._find_store_by_bucket(
            parsed_uri, location_map, 's3')

        self.assertEqual(store_name, 'store1')
        self.assertEqual(store_instance, store1)

        # Test with non-matching bucket
        parsed_uri = urlparse(
            's3://key:secret@s3.amazonaws.com/unknown-bucket/object')
        result = store_utils._find_store_by_bucket(
            parsed_uri, location_map, 's3')
        self.assertIsNone(result)

    @mock.patch('glance.common.store_utils._get_store_id_from_uri')
    @mock.patch('glance.common.store_utils.store_api')
    @mock.patch('glance.common.store_utils.CONF')
    def test_credential_rotation_with_multi_store_migration(
            self, mock_conf, mock_store_api, mock_get_store_id):
        mock_conf.enabled_backends = ['store1', 'store2']
        # No store found by URI (legacy behavior)
        mock_get_store_id.return_value = None
        # Simulate the scenario: multi-store config with different credentials
        stores = {
            'store1': {
                'access_key': 'new_key1',
                'secret_key': 'new_secret1',
                'bucket': 'bucket1',
                's3_host': 's3.amazonaws.com'
            },
            'store2': {
                'access_key': 'new_key2',
                'secret_key': 'new_secret2',
                'bucket': 'bucket2',
                's3_host': 's3.amazonaws.com'
            },
        }
        mock_store_api.location.SCHEME_TO_CLS_BACKEND_MAP = {
            's3': self._mock_store_config(stores)
        }

        # Legacy image with old credentials and no store metadata
        location = self._create_s3_location(
            's3://old_key:old_secret@s3.amazonaws.com/bucket1/object')
        self.image.locations = [location]

        store_utils.update_store_in_locations(
            self.context, self.image, self.image_repo)

        self.assertEqual(location['metadata']['store'], 'store1')
        self.assertEqual(
            location['url'],
            's3://new_key1:new_secret1@s3.amazonaws.com/bucket1/object')
        self.image_repo.save.assert_called_once_with(self.image)



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base test classes for running non-stubbed tests (functional tests)

This module provides base classes for functional testing including
SynchronousAPIBase for synchronous API testing and MinimalIPv6TestBase
for IPv6 network testing.
"""

import abc
import atexit
import datetime
import os
import shutil
import signal
import socket
import sys
import tempfile
from testtools import content as ttc
import textwrap
import threading
import time
from unittest import mock
import urllib.parse as urlparse
import uuid

import fixtures
import glance_store
from oslo_config import cfg
from oslo_serialization import jsonutils
import webob

from glance.common import config
from glance.common import utils
from glance.common import wsgi
from glance.db.sqlalchemy import api as db_api
from glance import housekeeping
from glance import tests as glance_tests
from glance.tests import utils as test_utils

execute, get_unused_port = test_utils.execute, test_utils.get_unused_port
tracecmd_osmap = {'Linux': 'strace', 'FreeBSD': 'truss'}

SQLITE_CONN_TEMPLATE = 'sqlite:////%s/tests.sqlite'
CONF = cfg.CONF


class BaseServer(metaclass=abc.ABCMeta):
    """
    Class used to easily manage starting and stopping
    a server during functional test runs.
    """
    def __init__(self, test_dir, port, sock=None):
        """
        Creates a new Server object.

        :param test_dir: The directory where all test stuff is kept.
        :param port: The port to start a server up on.
        """
        self.debug = True
        self.no_venv = False
        self.test_dir = test_dir
        self.bind_port = port
        self.conf_file_name = None
        self.conf_base = None
        self.paste_conf_base = None
        self.exec_env = None
        self.deployment_flavor = ''
        self.show_image_direct_url = False
        self.show_multiple_locations = False
        self.do_secure_hash = True
        self.http_retries = '3'
        self.property_protection_file = ''
        self.needs_database = False
        self.log_file = None
        self.sock = sock
        self.fork_socket = True
        self.process_pid = None
        self.server_module = None
        self.stop_kill = False

    def write_conf(self, **kwargs):
        """
        Writes the configuration file for the server to its intended
        destination.  Returns the name of the configuration file and
        the over-ridden config content (may be useful for populating
        error messages).
        """
        if not self.conf_base:
            raise RuntimeError("Subclass did not populate config_base!")

        conf_override = self.__dict__.copy()
        if kwargs:
            conf_override.update(**kwargs)

        # A config file and paste.ini to use just for this test...we don't want
        # to trample on currently-running Glance servers, now do we?

        conf_dir = os.path.join(self.test_dir, 'etc')
        conf_filepath = os.path.join(conf_dir, "%s.conf" % self.server_name)
        if os.path.exists(conf_filepath):
            os.unlink(conf_filepath)
        paste_conf_filepath = conf_filepath.replace(".conf", "-paste.ini")
        if os.path.exists(paste_conf_filepath):
            os.unlink(paste_conf_filepath)
        utils.safe_mkdirs(conf_dir)

        def override_conf(filepath, overridden):
            with open(filepath, 'w') as conf_file:
                conf_file.write(overridden)
                conf_file.flush()
                return conf_file.name

        overridden_core = self.conf_base % conf_override
        self.conf_file_name = override_conf(conf_filepath, overridden_core)

        overridden_paste = ''
        if self.paste_conf_base:
            overridden_paste = self.paste_conf_base % conf_override
            override_conf(paste_conf_filepath, overridden_paste)

        overridden = ('==Core config==\n%s\n==Paste config==\n%s' %
                      (overridden_core, overridden_paste))

        return self.conf_file_name, overridden

    @abc.abstractmethod
    def start(self, expect_exit=True, expected_exitcode=0, **kwargs):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    def reload(self, expect_exit=True, expected_exitcode=0, **kwargs):
        """
        Start and stop the service to reload

        Any kwargs passed to this method will override the configuration
        value in the conf file used in starting the servers.
        """
        self.stop()
        return self.start(expect_exit=expect_exit,
                          expected_exitcode=expected_exitcode, **kwargs)

    def create_database(self):
        """Create database if required for this server"""
        if self.needs_database:
            conf_dir = os.path.join(self.test_dir, 'etc')
            utils.safe_mkdirs(conf_dir)
            conf_filepath = os.path.join(conf_dir, 'glance-manage.conf')

            with open(conf_filepath, 'w') as conf_file:
                conf_file.write('[database]\n')
                conf_file.write('connection = %s' % self.sql_connection)
                conf_file.flush()

            glance_db_env = 'GLANCE_DB_TEST_SQLITE_FILE'
            if glance_db_env in os.environ:
                # use the empty db created and cached as a tempfile
                # instead of spending the time creating a new one
                db_location = os.environ[glance_db_env]
                shutil.copyfile(db_location, "%s/tests.sqlite" % self.test_dir)
            else:
                cmd = ('%s -m glance.cmd.manage --config-file %s db sync' %
                       (sys.executable, conf_filepath))
                execute(cmd, no_venv=self.no_venv, exec_env=self.exec_env,
                        expect_exit=True)

                # copy the clean db to a temp location so that it
                # can be reused for future tests
                (osf, db_location) = tempfile.mkstemp()
                os.close(osf)
                shutil.copyfile('%s/tests.sqlite' % self.test_dir, db_location)
                os.environ[glance_db_env] = db_location

                # cleanup the temp file when the test suite is
                # complete
                def _delete_cached_db():
                    try:
                        os.remove(os.environ[glance_db_env])
                    except Exception:
                        glance_tests.logger.exception(
                            "Error cleaning up the file %s" %
                            os.environ[glance_db_env])
                atexit.register(_delete_cached_db)

    def dump_log(self):
        if not self.log_file:
            return "log_file not set for {name}".format(name=self.server_name)
        elif not os.path.exists(self.log_file):
            return "{log_file} for {name} did not exist".format(
                log_file=self.log_file, name=self.server_name)
        with open(self.log_file, 'r') as fptr:
            return fptr.read().strip()


class PosixServer(BaseServer):
    def start(self, expect_exit=True, expected_exitcode=0, **kwargs):
        """
        Starts the server.

        Any kwargs passed to this method will override the configuration
        value in the conf file used in starting the servers.
        """

        # Ensure the configuration file is written
        self.write_conf(**kwargs)

        self.create_database()

        cmd = ("%(server_module)s --config-file %(conf_file_name)s"
               % {"server_module": self.server_module,
                  "conf_file_name": self.conf_file_name})
        cmd = "%s -m %s" % (sys.executable, cmd)
        # close the sock and release the unused port closer to start time
        if self.exec_env:
            exec_env = self.exec_env.copy()
        else:
            exec_env = {}
        pass_fds = set()
        if self.sock:
            if not self.fork_socket:
                self.sock.close()
                self.sock = None
            else:
                fd = os.dup(self.sock.fileno())
                exec_env[utils.GLANCE_TEST_SOCKET_FD_STR] = str(fd)
                pass_fds.add(fd)
                self.sock.close()

        self.process_pid = test_utils.fork_exec(cmd,
                                                logfile=os.devnull,
                                                exec_env=exec_env,
                                                pass_fds=pass_fds)

        self.stop_kill = not expect_exit
        if self.pid_file:
            pf = open(self.pid_file, 'w')
            pf.write('%d\n' % self.process_pid)
            pf.close()
        if not expect_exit:
            rc = 0
            try:
                os.kill(self.process_pid, 0)
            except OSError:
                raise RuntimeError("The process did not start")
        else:
            rc = test_utils.wait_for_fork(
                self.process_pid,
                expected_exitcode=expected_exitcode,
                force=False)
        # avoid an FD leak
        if self.sock:
            os.close(fd)
            self.sock = None
        return (rc, '', '')

    def stop(self):
        """
        Spin down the server.
        """
        if not self.process_pid:
            raise Exception('why is this being called? %s' % self.server_name)

        if self.stop_kill:
            os.kill(self.process_pid, signal.SIGTERM)
        rc = test_utils.wait_for_fork(self.process_pid, raise_error=False,
                                      force=self.stop_kill)
        return (rc, '', '')


Server = PosixServer


class ApiServer(Server):

    """
    Server object that starts/stops/manages the API server
    """

    def __init__(self, test_dir, port, policy_file, delayed_delete=False,
                 pid_file=None, sock=None, **kwargs):
        super(ApiServer, self).__init__(test_dir, port, sock=sock)
        self.server_name = 'api'
        self.server_module = 'glance.cmd.%s' % self.server_name
        self.default_store = kwargs.get("default_store", "file")
        self.bind_host = "127.0.0.1"
        self.metadata_encryption_key = "012345678901234567890123456789ab"
        self.image_dir = os.path.join(self.test_dir, "images")
        self.pid_file = pid_file or os.path.join(self.test_dir, "api.pid")
        self.log_file = os.path.join(self.test_dir, "api.log")
        self.image_size_cap = 1099511627776
        self.delayed_delete = delayed_delete
        self.workers = 0
        self.image_cache_dir = os.path.join(self.test_dir,
                                            'cache')
        self.image_cache_driver = 'sqlite'
        self.policy_file = policy_file
        self.policy_default_rule = 'default'
        self.property_protection_rule_format = 'roles'
        self.image_member_quota = 10
        self.image_property_quota = 10
        self.image_tag_quota = 10
        self.image_location_quota = 2
        self.disable_path = None

        self.enforce_new_defaults = True

        self.needs_database = True
        default_sql_connection = SQLITE_CONN_TEMPLATE % self.test_dir
        self.sql_connection = os.environ.get('GLANCE_TEST_SQL_CONNECTION',
                                             default_sql_connection)
        self.user_storage_quota = '0'
        self.lock_path = self.test_dir

        self.node_staging_uri = 'file://%s' % os.path.join(
            self.test_dir, 'staging')

        self.conf_base = """[DEFAULT]
debug = %(debug)s
default_log_levels = stevedore.extension=INFO
bind_host = %(bind_host)s
bind_port = %(bind_port)s
metadata_encryption_key = %(metadata_encryption_key)s
log_file = %(log_file)s
image_size_cap = %(image_size_cap)d
delayed_delete = %(delayed_delete)s
workers = %(workers)s
image_cache_dir = %(image_cache_dir)s
image_cache_driver = %(image_cache_driver)s
show_image_direct_url = %(show_image_direct_url)s
show_multiple_locations = %(show_multiple_locations)s
do_secure_hash = %(do_secure_hash)s
http_retries = %(http_retries)s
user_storage_quota = %(user_storage_quota)s
lock_path = %(lock_path)s
property_protection_file = %(property_protection_file)s
property_protection_rule_format = %(property_protection_rule_format)s
image_member_quota=%(image_member_quota)s
image_property_quota=%(image_property_quota)s
image_tag_quota=%(image_tag_quota)s
image_location_quota=%(image_location_quota)s
node_staging_uri=%(node_staging_uri)s
[database]
connection = %(sql_connection)s
[oslo_policy]
policy_file = %(policy_file)s
policy_default_rule = %(policy_default_rule)s
enforce_new_defaults=%(enforce_new_defaults)s
[paste_deploy]
flavor = %(deployment_flavor)s
[glance_store]
filesystem_store_datadir=%(image_dir)s
default_store = %(default_store)s
[import_filtering_opts]
allowed_ports = []
"""
        self.paste_conf_base = """[composite:glance-api]
paste.composite_factory = glance.api:root_app_factory
/: api
/healthcheck: healthcheck

[pipeline:api]
pipeline =
    cors
    versionnegotiation
    gzip
    unauthenticated-context
    rootapp

[composite:glance-api-caching]
paste.composite_factory = glance.api:root_app_factory
/: api-caching
/healthcheck: healthcheck

[pipeline:api-caching]
pipeline = cors versionnegotiation gzip context cache rootapp

[composite:glance-api-cachemanagement]
paste.composite_factory = glance.api:root_app_factory
/: api-cachemanagement
/healthcheck: healthcheck

[pipeline:api-cachemanagement]
pipeline =
    cors
    versionnegotiation
    gzip
    unauthenticated-context
    cache
    cache_manage
    rootapp

[composite:glance-api-fakeauth]
paste.composite_factory = glance.api:root_app_factory
/: api-fakeauth
/healthcheck: healthcheck

[pipeline:api-fakeauth]
pipeline = cors versionnegotiation gzip fakeauth context rootapp

[composite:glance-api-noauth]
paste.composite_factory = glance.api:root_app_factory
/: api-noauth
/healthcheck: healthcheck

[pipeline:api-noauth]
pipeline = cors versionnegotiation gzip context rootapp

[composite:rootapp]
paste.composite_factory = glance.api:root_app_factory
/: apiversions
/v2: apiv2app

[app:apiversions]
paste.app_factory = glance.api.versions:create_resource

[app:apiv2app]
paste.app_factory = glance.api.v2.router:API.factory

[app:healthcheck]
paste.app_factory = oslo_middleware:Healthcheck.app_factory
backends = disable_by_file
disable_by_file_path = %(disable_path)s

[filter:versionnegotiation]
paste.filter_factory = glance.api.middleware.version_negotiation:VersionNegotiationFilter.factory

[filter:gzip]
paste.filter_factory = glance.api.middleware.gzip:GzipMiddleware.factory

[filter:cache]
paste.filter_factory = glance.api.middleware.cache:CacheFilter.factory

[filter:cache_manage]
paste.filter_factory = glance.api.middleware.cache_manage:CacheManageFilter.factory

[filter:context]
paste.filter_factory = glance.api.middleware.context:ContextMiddleware.factory

[filter:unauthenticated-context]
paste.filter_factory = glance.api.middleware.context:UnauthenticatedContextMiddleware.factory

[filter:fakeauth]
paste.filter_factory = glance.tests.utils:FakeAuthMiddleware.factory

[filter:cors]
paste.filter_factory = oslo_middleware.cors:filter_factory
allowed_origin=http://valid.example.com
"""


class MinimalIPv6TestBase(test_utils.BaseTestCase):
    """Minimal test base class for IPv6 network testing.

    This class provides only the essential functionality needed to test
    IPv6 network binding and communication. It starts a separate server
    process for testing IPv6 network functionality.
    """

    def setUp(self):
        super(MinimalIPv6TestBase, self).setUp()
        self.test_dir = self.useFixture(fixtures.TempDir()).path

        self.api_port, api_sock = test_utils.get_unused_port_and_socket()

        self.config(bind_host='::1')
        self.config(image_cache_dir=self.test_dir)

        conf_dir = os.path.join(self.test_dir, 'etc')
        utils.safe_mkdirs(conf_dir)
        self.copy_data_file('schema-image.json', conf_dir)
        self.policy_file = os.path.join(conf_dir, 'policy.yaml')

        with open(self.policy_file, 'w') as f:
            f.write('{}')

        self.api_server = ApiServer(self.test_dir,
                                    self.api_port,
                                    self.policy_file,
                                    sock=api_sock)
        self.api_server.bind_host = '::1'

        self.launched_servers = []
        self.files_to_destroy = []

        self.addCleanup(
            self._reset_database, self.api_server.sql_connection)
        self.addCleanup(self.cleanup)
        self._reset_database(self.api_server.sql_connection)

    def copy_data_file(self, file_name, dst_dir):
        src_file_name = os.path.join('glance/tests/etc', file_name)
        shutil.copy(src_file_name, dst_dir)
        dst_file_name = os.path.join(dst_dir, file_name)
        return dst_file_name

    def _reset_database(self, conn_string):
        """Reset the database for clean test state."""
        if conn_string.startswith('sqlite'):
            pass
        elif conn_string.startswith('mysql'):
            conn_pieces = urlparse.urlparse(conn_string)
            database = conn_pieces.path.strip('/')
            loc_pieces = conn_pieces.netloc.split('@')
            host = loc_pieces[1]
            auth_pieces = loc_pieces[0].split(':')
            user = auth_pieces[0]
            password = ""
            if len(auth_pieces) > 1:
                if auth_pieces[1].strip():
                    password = "-p%s" % auth_pieces[1]
            sql = ("drop database if exists %(database)s; "
                   "create database %(database)s;") % {'database': database}
            cmd = ("mysql -u%(user)s %(password)s -h%(host)s "
                   "-e\"%(sql)s\"") % {'user': user, 'password': password,
                                       'host': host, 'sql': sql}
            exitcode, out, err = execute(cmd)
            self.assertEqual(0, exitcode)

    def cleanup(self):
        """Stop any started servers and clean up files."""
        servers = [self.api_server]
        for s in servers:
            try:
                s.stop()
            except Exception:
                pass

        for f in self.files_to_destroy:
            if os.path.exists(f):
                os.unlink(f)

    def ping_server_ipv6(self, port):
        """Ping server on IPv6 port."""
        s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
        try:
            s.connect(("::1", port))
            return True
        except socket.error:
            return False
        finally:
            s.close()

    def wait_for_servers(self, servers, expect_launch=True, timeout=30):
        """Wait for servers to be ready on IPv6."""
        now = datetime.datetime.now()
        timeout_time = now + datetime.timedelta(seconds=timeout)
        replied = []
        while (timeout_time > now):
            pinged = 0
            for server in servers:
                if self.ping_server_ipv6(server.bind_port):
                    pinged += 1
                    if server not in replied:
                        replied.append(server)
            if pinged == len(servers):
                return None if expect_launch else 'Unexpected server launch'
            now = datetime.datetime.now()
            time.sleep(0.05)

        failed = list(set(servers) - set(replied))
        if failed:
            return 'Servers failed to start: %s' % [s.server_name for s in failed]
        return None if expect_launch else 'Unexpected server launch'

    def start_with_retry(self, server, port_name, max_retries,
                         expect_launch=True, **kwargs):
        """Start server with retry logic."""
        launch_msg = None
        for i in range(max_retries):
            exitcode, out, err = server.start(expect_exit=not expect_launch,
                                              **kwargs)
            name = server.server_name
            self.assertEqual(0, exitcode,
                             "Failed to spin up the %s server. "
                             "Got: %s" % (name, err))
            launch_msg = self.wait_for_servers([server], expect_launch)
            if launch_msg:
                server.stop()
                port, sock = test_utils.get_unused_port_and_socket()
                server.bind_port = port
                if sock:
                    server.sock = sock
                setattr(self, port_name, port)
            else:
                self.launched_servers.append(server)
                break
        self.assertTrue(launch_msg is None, launch_msg)

    def start_servers(self, **kwargs):
        """Start the API server."""
        self.cleanup()
        self.start_with_retry(self.api_server, 'api_port', 3, **kwargs)


class SynchronousAPIBase(test_utils.BaseTestCase):
    """A base class that provides synchronous calling into the API.

    This provides a way to directly call into the API WSGI stack
    without starting a separate server, and with a simple paste
    pipeline. Configured with multi-store and a real database.

    This differs from other functional test approaches in that it calls
    directly into the WSGI stack rather than starting a separate server
    process. This test base is appropriate for situations where you
    need to be able to mock the state of the world (i.e. warp time, or
    inject errors) but should not be used for happy-path testing where
    a separate server process provides more isolation.

    To use this, inherit and run start_server() before you are ready
    to make API calls (either in your setUp() or per-test if you need
    to change config or mocking).

    Once started, use the api_get(), api_put(), api_post(), and
    api_delete() methods to make calls to the API.

    """

    TENANT = str(uuid.uuid4())

    @mock.patch('oslo_db.sqlalchemy.enginefacade.writer.get_engine')
    def setup_database(self, mock_get_engine):
        """Configure and prepare a fresh sqlite database."""
        db_file = 'sqlite:///%s/test.db' % self.test_dir
        self.config(connection=db_file, group='database')

        # NOTE(danms): Make sure that we clear the current global
        # database configuration, provision a temporary database file,
        # and run migrations with our configuration to define the
        # schema there.
        db_api.clear_db_env()
        engine = db_api.get_engine()
        mock_get_engine.return_value = engine
        with mock.patch('logging.config'):
            # NOTE(danms): The alembic config in the env module will break our
            # BaseTestCase logging setup. So mock that out to prevent it while
            # we db_sync.
            test_utils.db_sync(engine=engine)

    def setup_scrubber_conf(self, daemon=False, wakeup_time=300,
                            scrub_pool_size=12):
        self.scrubber_conf = os.path.join(
            self.test_dir, 'glance-scrubber.conf')
        db_file = 'sqlite:///%s/test.db' % self.test_dir
        with open(self.scrubber_conf, 'w') as f:
            f.write(textwrap.dedent("""
            [DEFAULT]
            enabled_backends=store1:file,store2:file,store3:file
            daemon=%(daemon)s
            wakeup_time=%(wakeup_time)s
            scrub_pool_size=%(scrub_pool_size)s
            [database]
            connection=%(connection)s
            [store1]
            filesystem_store_datadir=%(store1)s
            [store2]
            filesystem_store_datadir=%(store2)s
            [store3]
            filesystem_store_datadir=%(store3)s
            [os_glance_staging_dir]
            filesystem_store_datadir=%(staging)s
            [glance_store]
            default_backend=store1
            """) % {
                "daemon": daemon,
                "wakeup_time": wakeup_time,
                "scrub_pool_size": scrub_pool_size,
                "connection": db_file,
                "store1": self._store_dir('store1'),
                "store2": self._store_dir('store2'),
                "store3": self._store_dir('store3'),
                "staging": self._store_dir('staging'),
            })

    def setup_simple_paste(self):
        """Setup a very simple no-auth paste pipeline.

        This configures the API to be very direct, including only the
        middleware absolutely required for consistent API calls.
        """
        self.paste_config = os.path.join(self.test_dir, 'glance-api-paste.ini')
        with open(self.paste_config, 'w') as f:
            f.write(textwrap.dedent("""
            [filter:context]
            paste.filter_factory = glance.api.middleware.context:\
                ContextMiddleware.factory
            [filter:unauthenticated-context]
            paste.filter_factory = glance.api.middleware.context:\
                UnauthenticatedContextMiddleware.factory
            [filter:versionnegotiation]
            paste.filter_factory = glance.api.middleware.version_negotiation:\
                VersionNegotiationFilter.factory
            [filter:fakeauth]
            paste.filter_factory = glance.tests.utils:\
                FakeAuthMiddleware.factory
            [filter:cors]
            paste.filter_factory = oslo_middleware.cors:filter_factory
            allowed_origin=http://valid.example.com
            [filter:gzip]
            paste.filter_factory = glance.api.middleware.gzip:\
                GzipMiddleware.factory
            [filter:cache]
            paste.filter_factory = glance.api.middleware.cache:\
            CacheFilter.factory
            [filter:cachemanage]
            paste.filter_factory = glance.api.middleware.cache_manage:\
            CacheManageFilter.factory
            [pipeline:glance-api-cachemanagement]
            pipeline = context cache cachemanage gzip rootapp
            [pipeline:glance-api-cors]
            pipeline = cors context gzip rootapp
            [pipeline:glance-api-caching]
            pipeline = context cache gzip rootapp
            [pipeline:glance-api]
            pipeline = context rootapp
            [pipeline:glance-api-versionnegotiation]
            pipeline = versionnegotiation unauthenticated-context cache cachemanage rootapp
            [pipeline:glance-api-fake]
            pipeline = fakeauth context gzip rootapp
            [composite:rootapp]
            paste.composite_factory = glance.api:root_app_factory
            /: apiversions
            /v2: apiv2app
            /healthcheck: healthcheck
            [app:apiversions]
            paste.app_factory = glance.api.versions:create_resource
            [app:apiv2app]
            paste.app_factory = glance.api.v2.router:API.factory
            [app:healthcheck]
            paste.app_factory = oslo_middleware:Healthcheck.app_factory
            backends = disable_by_file
            disable_by_file_path = /tmp/test_path
            """))

    def _store_dir(self, store):
        return os.path.join(self.test_dir, store)

    def setup_single_store(self):
        """Configures single backend.

        This configures the API with one file-backed store
        as well as node_staging_uri for imports.

        """
        glance_store.register_opts(CONF)
        self.config(filesystem_store_datadir=self._store_dir('store1'),
                    group='glance_store')
        node_staging_uri = 'file://%s' % os.path.join(
            self.test_dir, 'staging')
        utils.safe_mkdirs(node_staging_uri[7:])
        self.config(node_staging_uri=node_staging_uri)
        self.config(default_store='file', group='glance_store')
        glance_store.create_stores(CONF)
        glance_store.verify_default_store()

    def setup_property_protection_files(self):
        """Setup property protection configuration files.

        This creates property protection files dynamically for testing,
        similar to how setup_simple_paste() creates paste config files.
        Creates both roles-based and policies-based property protection
        files.
        """
        conf_dir = os.path.join(self.test_dir, 'etc')
        utils.safe_mkdirs(conf_dir)

        # Create roles-based property protection file
        property_file_roles = os.path.join(conf_dir,
                                           'property-protections.conf')
        with open(property_file_roles, 'w') as f:
            f.write(textwrap.dedent("""
            [^x_owner_.*]
            create = admin,member
            read = admin,member
            update = admin,member
            delete = admin,member

            [spl_create_prop]
            create = admin,spl_role
            read = admin,spl_role
            update = admin
            delete = admin

            [spl_read_prop]
            create = admin,spl_role
            read = admin,spl_role
            update = admin
            delete = admin

            [spl_read_only_prop]
            create = admin
            read = admin,spl_role
            update = admin
            delete = admin

            [spl_update_prop]
            create = admin,spl_role
            read = admin,spl_role
            update = admin,spl_role
            delete = admin

            [spl_update_only_prop]
            create = admin
            read = admin
            update = admin,spl_role
            delete = admin

            [spl_delete_prop]
            create = admin,spl_role
            read = admin,spl_role
            update = admin
            delete = admin,spl_role

            [spl_delete_empty_prop]
            create = admin,spl_role
            read = admin,spl_role
            update = admin
            delete = admin,spl_role

            [^x_all_permitted.*]
            create = @
            read = @
            update = @
            delete = @

            [^x_none_permitted.*]
            create = !
            read = !
            update = !
            delete = !

            [x_none_read]
            create = admin,member
            read = !
            update = !
            delete = !

            [x_none_update]
            create = admin,member
            read = admin,member
            update = !
            delete = admin,member

            [x_none_delete]
            create = admin,member
            read = admin,member
            update = admin,member
            delete = !

            [x_case_insensitive]
            create = admin,Member
            read = admin,Member
            update = admin,Member
            delete = admin,Member

            [x_foo_matcher]
            create = admin
            read = admin
            update = admin
            delete = admin

            [x_foo_*]
            create = @
            read = @
            update = @
            delete = @

            [.*]
            create = admin
            read = admin
            update = admin
            delete = admin
            """))

        # Create policies-based property protection file
        property_file_policies = os.path.join(
            conf_dir, 'property-protections-policies.conf')
        with open(property_file_policies, 'w') as f:
            f.write(textwrap.dedent("""
            [spl_creator_policy]
            create = glance_creator
            read = glance_creator
            update = context_is_admin
            delete = context_is_admin

            [spl_default_policy]
            create = context_is_admin
            read = default
            update = context_is_admin
            delete = context_is_admin

            [^x_all_permitted.*]
            create = @
            read = @
            update = @
            delete = @

            [^x_none_permitted.*]
            create = !
            read = !
            update = !
            delete = !

            [x_none_read]
            create = context_is_admin
            read = !
            update = !
            delete = !

            [x_none_update]
            create = context_is_admin
            read = context_is_admin
            update = !
            delete = context_is_admin

            [x_none_delete]
            create = context_is_admin
            read = context_is_admin
            update = context_is_admin
            delete = !

            [x_foo_matcher]
            create = context_is_admin
            read = context_is_admin
            update = context_is_admin
            delete = context_is_admin

            [x_foo_*]
            create = @
            read = @
            update = @
            delete = @

            [.*]
            create = context_is_admin
            read = context_is_admin
            update = context_is_admin
            delete = context_is_admin
            """))

        self.property_file_roles = property_file_roles
        self.property_file_policies = property_file_policies

    def setup_stores(self):
        """Configures multiple backend stores.

        This configures the API with three file-backed stores (store1,
        store2, and store3) as well as a os_glance_staging_store for
        imports.

        """
        self.config(enabled_backends={'store1': 'file', 'store2': 'file',
                                      'store3': 'file'})
        glance_store.register_store_opts(CONF,
                                         reserved_stores=wsgi.RESERVED_STORES)
        self.config(default_backend='store1',
                    group='glance_store')
        self.config(filesystem_store_datadir=self._store_dir('store1'),
                    group='store1')
        self.config(filesystem_store_datadir=self._store_dir('store2'),
                    group='store2')
        self.config(filesystem_store_datadir=self._store_dir('store3'),
                    group='store3')
        self.config(filesystem_store_datadir=self._store_dir('staging'),
                    group='os_glance_staging_store')
        self.config(filesystem_store_datadir=self._store_dir('tasks'),
                    group='os_glance_tasks_store')

        glance_store.create_multi_stores(CONF,
                                         reserved_stores=wsgi.RESERVED_STORES)
        glance_store.verify_store()

    def setUp(self, single_store=False, bypass_headers=False):
        super(SynchronousAPIBase, self).setUp()
        self.bypass_headers = bypass_headers

        self.setup_database()
        self.setup_simple_paste()
        if single_store:
            self.setup_single_store()
        else:
            self.setup_stores()

        conf_dir = os.path.join(self.test_dir, 'etc')
        utils.safe_mkdirs(conf_dir)
        self.copy_data_file('schema-image.json', conf_dir)
        self.setup_property_protection_files()

    def copy_data_file(self, file_name, dst_dir):
        src_file_name = os.path.join('glance/tests/etc', file_name)
        shutil.copy(src_file_name, dst_dir)
        dst_file_name = os.path.join(dst_dir, file_name)
        return dst_file_name

    def start_server(self, enable_cache=True, set_worker_url=True,
                     use_fake_auth=False, run_staging_cleaner=False,
                     enable_cors=False, enable_version_negotiation=False):
        """Builds and "starts" the API server.

        Note that this doesn't actually start a separate server process,
        but instead builds the WSGI application directly. The terminology
        is used here to maintain consistency with other test patterns.
        """
        config.set_config_defaults()
        root_app = 'glance-api'
        if enable_version_negotiation:
            root_app = 'glance-api-versionnegotiation'
            self.config(image_cache_dir=self._store_dir('cache'))
        elif enable_cache:
            root_app = 'glance-api-cachemanagement'
            self.config(image_cache_dir=self._store_dir('cache'))

        if enable_cors:
            root_app = 'glance-api-cors'

        if use_fake_auth:
            root_app = 'glance-api-fake'

        if set_worker_url:
            self.config(worker_self_reference_url='http://workerx')

        if run_staging_cleaner:
            cleaner = housekeeping.StagingStoreCleaner(db_api)
            # NOTE(danms): Start thread as a daemon. It is still a
            # single-shot, but this will not block our shutdown if it is
            # running.
            cleanup_thread = threading.Thread(
                target=cleaner.clean_orphaned_staging_residue,
                daemon=True)
            cleanup_thread.start()

        self.api = config.load_paste_app(root_app,
                                         conf_file=self.paste_config)
        self.config(enforce_new_defaults=True,
                    group='oslo_policy')
        self.config(enforce_scope=True,
                    group='oslo_policy')

    def start_scrubber(self, daemon=False, wakeup_time=300,
                       restore=None, raise_error=True):
        self.setup_scrubber_conf(daemon=daemon, wakeup_time=wakeup_time)
        exe_cmd = f"{sys.executable} -m glance.cmd.scrubber"

        # Modify command based on restore and daemon flags
        if restore:
            exe_cmd += f" --restore {restore}"
        if daemon:
            exe_cmd += " --daemon"

        # Prepare the final command string with the config directory
        cmd = f"{exe_cmd} --config-dir {self.test_dir}"

        # Determine if we need to return the process object
        expect_exit = not daemon
        return_process = daemon

        # Execute the command and return the result
        return execute(cmd, raise_error=raise_error, expect_exit=expect_exit,
                       return_process=return_process)

    def _headers(self, custom_headers=None):
        base_headers = {
            'X-Identity-Status': 'Confirmed',
            'X-Auth-Token': '932c5c84-02ac-4fe5-a9ba-620af0e2bb96',
            'X-User-Id': 'f9a41d13-0c13-47e9-bee2-ce4e8bfe958e',
            'X-Tenant-Id': self.TENANT,
            'Content-Type': 'application/json',
            'X-Roles': 'admin',
        }
        base_headers.update(custom_headers or {})
        return base_headers

    def api_request(self, method, url, headers=None, data=None,
                    json=None, body_file=None):
        """Perform a request against the API.

        NOTE: Most code should use api_get(), api_post(), api_put(),
              or api_delete() instead!

        :param method: The HTTP method to use (i.e. GET, POST, etc)
        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param headers: Optional updates to the default set of headers
        :param data: Optional bytes data payload to send (overrides @json)
        :param json: Optional dict structure to be jsonified and sent as
                     the payload (mutually exclusive with @data)
        :param body_file: Optional io.IOBase to provide as the input data
                          stream for the request (overrides @data)
        :returns: A webob.Response object
        """
        if not self.bypass_headers:
            headers = self._headers(headers)
        else:
            # Ensure headers is at least an empty dict, not None
            if headers is None:
                headers = {}
        # Set a base URL for the request so application_url is available
        # Use host_url to ensure application_url is properly set
        req = webob.Request.blank(url, method=method,
                                  headers=headers)
        if not req.application_url:
            req.environ['wsgi.url_scheme'] = 'http'
            req.environ['HTTP_HOST'] = 'localhost'
            req.environ['SERVER_NAME'] = 'localhost'
            req.environ['SERVER_PORT'] = '80'
        if json and not data:
            data = jsonutils.dumps(json).encode()
        if data and not body_file:
            req.body = data
        elif body_file:
            req.body_file = body_file
        return self._call_api(req)

    def api_get(self, url, headers=None, data=None, json=None):
        """Perform a GET request against the API.

        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param headers: Optional updates to the default set of headers
        :param data: Optional bytes data payload to send (overrides @json)
        :param json: Optional dict structure to be jsonified and sent as
                     the payload (mutually exclusive with @data)
        :returns: A webob.Response object
        """
        return self.api_request('GET', url, headers=headers,
                                data=data, json=json)

    def api_post(self, url, headers=None, data=None, json=None,
                 body_file=None):
        """Perform a POST request against the API.

        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param headers: Optional updates to the default set of headers
        :param data: Optional bytes data payload to send (overrides @json)
        :param json: Optional dict structure to be jsonified and sent as
                     the payload (mutually exclusive with @data)
        :param body_file: Optional io.IOBase to provide as the input data
                          stream for the request (overrides @data)
        :returns: A webob.Response object
        """
        return self.api_request('POST', url, headers=headers,
                                data=data, json=json,
                                body_file=body_file)

    def api_put(self, url, headers=None, data=None, json=None, body_file=None):
        """Perform a PUT request against the API.

        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param headers: Optional updates to the default set of headers
        :param data: Optional bytes data payload to send (overrides @json,
                     mutually exclusive with body_file)
        :param json: Optional dict structure to be jsonified and sent as
                     the payload (mutually exclusive with @data)
        :param body_file: Optional io.IOBase to provide as the input data
                          stream for the request (overrides @data)
        :returns: A webob.Response object
        """
        return self.api_request('PUT', url, headers=headers,
                                data=data, json=json,
                                body_file=body_file)

    def api_delete(self, url, headers=None, data=None, json=None):
        """Perform a DELETE request against the API.

        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param headers: Optional updates to the default set of headers
        :param data: Optional bytes data payload to send (overrides @json)
        :param json: Optional dict structure to be jsonified and sent as
                     the payload (mutually exclusive with @data)
        :returns: A webob.Response object
        """
        return self.api_request('DELETE', url, headers=headers,
                                data=data, json=json)

    def _call_api(self, req):
        """Call the WSGI application and handle Controller responses.

        This wraps the WSGI application call and handles cases where the
        version negotiation middleware returns a Controller object instead
        of a Response.

        :param req: A webob.Request object
        :returns: A webob.Response object
        """
        response = self.api(req)
        # The version negotiation filter may return a Controller for
        # unknown versions, which needs to be called to get the Response
        if hasattr(response, '__call__') and not hasattr(
                response, 'status_code'):
            response = response(req)
        return response

    def api_patch(self, url, patches, headers=None):
        """Perform a PATCH request against the API.

        :param url: The *path* part of the URL to call (i.e. /v2/images)
        :param patches: One or more patch dicts
        :param headers: Optional updates to the default set of headers
        :returns: A webob.Response object
        """
        json = patches if isinstance(patches, list) else list(patches)

        if not headers:
            headers = {}
        headers['Content-Type'] = \
            'application/openstack-images-v2.1-json-patch'
        return self.api_request('PATCH', url, headers=headers,
                                json=json)

    def _import_copy(self, image_id, stores, headers=None):
        """Do an import of image_id to the given stores."""
        body = {'method': {'name': 'copy-image'},
                'stores': stores,
                'all_stores': False}

        return self.api_post(
            '/v2/images/%s/import' % image_id,
            headers=headers,
            json=body)

    def _import_direct(self, image_id, stores, headers=None):
        """Do an import of image_id to the given stores."""
        body = {'method': {'name': 'glance-direct'},
                'stores': stores,
                'all_stores': False}

        return self.api_post(
            '/v2/images/%s/import' % image_id,
            headers=headers,
            json=body)

    def _import_web_download(self, image_id, stores, url, headers=None):
        """Do an import of image_id to the given stores."""
        body = {'method': {'name': 'web-download',
                           'uri': url},
                'stores': stores,
                'all_stores': False}

        return self.api_post(
            '/v2/images/%s/import' % image_id,
            headers=headers,
            json=body)

    def _create_and_upload(self, data_iter=None, expected_code=204,
                           visibility=None):
        data = {
            'name': 'foo',
            'container_format': 'bare',
            'disk_format': 'raw'
        }
        if visibility:
            data['visibility'] = visibility

        resp = self.api_post('/v2/images',
                             json=data)
        self.assertEqual(201, resp.status_code, resp.text)
        image = jsonutils.loads(resp.text)

        if data_iter:
            resp = self.api_put(
                '/v2/images/%s/file' % image['id'],
                headers={'Content-Type': 'application/octet-stream'},
                body_file=data_iter)
        else:
            resp = self.api_put(
                '/v2/images/%s/file' % image['id'],
                headers={'Content-Type': 'application/octet-stream'},
                data=b'IMAGEDATA')
        self.assertEqual(expected_code, resp.status_code)

        return image['id']

    def _create_and_stage(self, data_iter=None, expected_code=204,
                          visibility=None, extra={}):
        data = {
            'name': 'foo',
            'container_format': 'bare',
            'disk_format': 'raw',
        }
        if visibility:
            data['visibility'] = visibility

        data.update(extra)
        resp = self.api_post('/v2/images',
                             json=data)
        image = jsonutils.loads(resp.text)

        if data_iter:
            resp = self.api_put(
                '/v2/images/%s/stage' % image['id'],
                headers={'Content-Type': 'application/octet-stream'},
                body_file=data_iter)
        else:
            resp = self.api_put(
                '/v2/images/%s/stage' % image['id'],
                headers={'Content-Type': 'application/octet-stream'},
                data=b'IMAGEDATA')
        self.assertEqual(expected_code, resp.status_code)

        return image['id']

    def _wait_for_import(self, image_id, retries=10):
        for i in range(0, retries):
            image = self.api_get('/v2/images/%s' % image_id).json
            if not image.get('os_glance_import_task'):
                break
            self.addDetail('Create-Import task id',
                           ttc.text_content(image['os_glance_import_task']))
            time.sleep(1)

        self.assertIsNone(image.get('os_glance_import_task'),
                          'Timed out waiting for task to complete')

        return image

    def _create_and_import(self, stores=[], data_iter=None, expected_code=202,
                           visibility=None, extra={}):
        """Create an image, stage data, and import into the given stores.

        :returns: image_id
        """
        image_id = self._create_and_stage(data_iter=data_iter,
                                          visibility=visibility,
                                          extra=extra)

        resp = self._import_direct(image_id, stores)
        self.assertEqual(expected_code, resp.status_code)

        if expected_code >= 400:
            return image_id

        # Make sure it becomes active
        image = self._wait_for_import(image_id)
        self.assertEqual('active', image['status'])

        return image_id

    def _get_latest_task(self, image_id):
        tasks = self.api_get('/v2/images/%s/tasks' % image_id).json['tasks']
        tasks = sorted(tasks, key=lambda t: t['updated_at'])
        self.assertGreater(len(tasks), 0)
        return tasks[-1]

    def _create(self):
        return self.api_post('/v2/images',
                             json={'name': 'foo',
                                   'container_format': 'bare',
                                   'disk_format': 'raw'})

    def _create_metadef_resource(self, path=None, data=None,
                                 expected_code=201):
        resp = self.api_post(path,
                             json=data)
        md_resource = jsonutils.loads(resp.text)
        self.assertEqual(expected_code, resp.status_code)
        return md_resource



# Copyright 2013 Red Hat, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests gzip middleware."""

from glance.tests import functional
from glance.tests import utils


class GzipMiddlewareTest(functional.SynchronousAPIBase):

    def setUp(self):
        super(GzipMiddlewareTest, self).setUp()
        self.start_server()

    @utils.skip_if_disabled
    def test_gzip_requests(self):
        # Accept-Encoding: Identity
        headers = self._headers({'Accept-Encoding': 'identity'})
        response = self.api_get('/v2/images', headers=headers)
        # When identity is requested, Content-Encoding should not be set
        self.assertIsNone(response.headers.get('content-encoding'))

        # Accept-Encoding: gzip
        headers = self._headers({'Accept-Encoding': 'gzip'})
        response = self.api_get('/v2/images', headers=headers)
        # When gzip is requested, Content-Encoding should be set to gzip
        self.assertEqual('gzip', response.headers.get('content-encoding'))



# Copyright 2011-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from keystoneauth1 import session
from keystoneauth1 import token_endpoint
from oslo_config import cfg
from oslo_context import context

from glance.api import policy

CONF = cfg.CONF


def get_ksa_client(context):
    """Returns a keystoneauth Adapter using token from context.

    This will return a simple keystoneauth adapter that can be used to
    make requests against a remote service using the token provided
    (and already authenticated) from the user and stored in a
    RequestContext.

    :param context: User request context
    :returns: keystoneauth1 Adapter object
    """
    auth = token_endpoint.Token(CONF.keystone_authtoken.identity_uri,
                                context.auth_token)
    return session.Session(auth=auth)


class RequestContext(context.RequestContext):
    """Stores information about the security context.

    Stores how the user accesses the system, as well as additional request
    information.

    """

    def __init__(self, service_catalog=None, policy_enforcer=None, **kwargs):
        super(RequestContext, self).__init__(**kwargs)
        self.service_catalog = service_catalog
        self.policy_enforcer = policy_enforcer or policy.Enforcer()
        if not self.is_admin:
            self.is_admin = self.policy_enforcer.check_is_admin(self)

    def to_dict(self):
        d = super(RequestContext, self).to_dict()
        d.update({
            'roles': self.roles,
            'service_catalog': self.service_catalog,
        })
        return d

    @classmethod
    def from_dict(cls, values):
        return cls(**values)

    @property
    def owner(self):
        """Return the owner to correlate with an image."""
        return self.project_id

    @property
    def can_see_deleted(self):
        """Admins can see deleted by default"""
        return self.show_deleted or self.is_admin

    def elevated(self):
        """Return a copy of this context with admin flag set."""

        context = copy.copy(self)
        context.roles = copy.deepcopy(self.roles)
        if 'admin' not in context.roles:
            context.roles.append('admin')

        context.is_admin = True

        return context


def get_admin_context(show_deleted=False):
    """Create an administrator context."""
    return RequestContext(auth_token=None,
                          project_id=None,
                          is_admin=True,
                          show_deleted=show_deleted,
                          overwrite=False)
